from __future__ import annotations

import asyncio
import logging
import traceback
from collections.abc import Coroutine
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, NoReturn, Optional, TypeVar

import typer
from click.exceptions import Exit
from httpx import HTTPError
from rich.console import Console
from rich.logging import RichHandler
from rich.markup import escape

from ..exceptions import (
    AuthenticationError,
    Error,
    FileNotValidError,
    GraphQLError,
    NodeNotFoundError,
    ResourceNotDefinedError,
    SchemaNotFoundError,
    ServerNotReachableError,
    ServerNotResponsiveError,
    ValidationError,
)
from ..yaml import YamlFile
from .client import initialize_client_sync
from .exceptions import QueryNotFoundError

if TYPE_CHECKING:
    from ..schema.repository import InfrahubRepositoryConfig
    from ..spec.object import ObjectFile

YamlFileVar = TypeVar("YamlFileVar", bound=YamlFile)
T = TypeVar("T")


def init_logging(debug: bool = False) -> None:
    logging.getLogger("infrahub_sdk").setLevel(logging.CRITICAL)
    logging.getLogger("httpx").setLevel(logging.ERROR)
    logging.getLogger("httpcore").setLevel(logging.ERROR)

    log_level = "DEBUG" if debug else "INFO"
    FORMAT = "%(message)s"
    logging.basicConfig(level=log_level, format=FORMAT, datefmt="[%X]", handlers=[RichHandler()])
    logging.getLogger("infrahubctl")


def handle_exception(exc: Exception, console: Console, exit_code: int) -> NoReturn:
    """Handle exeception in a different fashion based on its type."""
    if isinstance(exc, Exit):
        raise typer.Exit(code=exc.exit_code)
    if isinstance(exc, AuthenticationError):
        console.print(f"[red]Authentication failure: {exc!s}")
        raise typer.Exit(code=exit_code)
    if isinstance(exc, (ServerNotReachableError, ServerNotResponsiveError)):
        console.print(f"[red]{exc!s}")
        raise typer.Exit(code=exit_code)
    if isinstance(exc, HTTPError):
        console.print(f"[red]HTTP communication failure: {exc!s} on {exc.request.method} to {exc.request.url}")
        raise typer.Exit(code=exit_code)
    if isinstance(exc, GraphQLError):
        print_graphql_errors(console=console, errors=exc.errors)
        raise typer.Exit(code=exit_code)
    if isinstance(exc, (SchemaNotFoundError, NodeNotFoundError, ResourceNotDefinedError)):
        console.print(f"[red]Error: {exc!s}")
        raise typer.Exit(code=exit_code)

    console.print(f"[red]Error: {exc!s}")
    console.print(traceback.format_exc())
    raise typer.Exit(code=exit_code)


def catch_exception(
    console: Console | None = None, exit_code: int = 1
) -> Callable[[Callable[..., T]], Callable[..., T | Coroutine[Any, Any, T]]]:
    """Decorator to handle exception for commands."""
    if not console:
        console = Console()

    def decorator(func: Callable[..., T]) -> Callable[..., T | Coroutine[Any, Any, T]]:
        if asyncio.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args: Any, **kwargs: Any) -> T:
                try:
                    return await func(*args, **kwargs)
                except (Error, Exception) as exc:
                    return handle_exception(exc=exc, console=console, exit_code=exit_code)

            return async_wrapper

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            try:
                return func(*args, **kwargs)
            except (Error, Exception) as exc:
                return handle_exception(exc=exc, console=console, exit_code=exit_code)

        return wrapper

    return decorator


def execute_graphql_query(
    query: str,
    variables_dict: dict[str, Any],
    repository_config: InfrahubRepositoryConfig,
    branch: str | None = None,
    debug: bool = False,
) -> dict:
    console = Console()
    query_object = repository_config.get_query(name=query)
    query_str = query_object.load_query()

    client = initialize_client_sync()
    response = client.execute_graphql(
        query=query_str,
        branch_name=branch,
        variables=variables_dict,
        raise_for_error=False,
    )

    if debug:
        console.print("-" * 40)
        console.print(f"Response for GraphQL Query {query}")
        console.print(response)
        console.print("-" * 40)

    return response


def print_graphql_errors(console: Console, errors: list) -> None:
    if not isinstance(errors, list):
        console.print(f"[red]{escape(str(errors))}")

    for error in errors:
        if isinstance(error, dict) and "message" in error and "path" in error:
            console.print(f"[red]{escape(str(error['path']))} {escape(str(error['message']))}")
        else:
            console.print(f"[red]{escape(str(error))}")


def parse_cli_vars(variables: Optional[list[str]]) -> dict[str, str]:
    if not variables:
        return {}

    return {var.split("=")[0]: var.split("=")[1] for var in variables if "=" in var}


def find_graphql_query(name: str, directory: str | Path = ".") -> str:
    if isinstance(directory, str):
        directory = Path(directory)

    for query_file in directory.glob("**/*.gql"):
        if query_file.stem != name:
            continue
        return query_file.read_text(encoding="utf-8")

    raise QueryNotFoundError(name=name)


def render_action_rich(value: str) -> str:
    if value == "created":
        return f"[green]{value.upper()}[/green]"
    if value == "updated":
        return f"[magenta]{value.upper()}[/magenta]"
    if value == "deleted":
        return f"[red]{value.upper()}[/red]"

    return value.upper()


def get_fixtures_dir() -> Path:
    """Get the directory which stores fixtures that are common to multiple unit/integration tests."""
    here = Path(__file__).resolve().parent
    return here.parent.parent / "tests" / "fixtures"


def load_yamlfile_from_disk_and_exit(
    paths: list[Path], file_type: type[YamlFileVar], console: Console
) -> list[YamlFileVar]:
    has_error = False
    try:
        data_files = file_type.load_from_disk(paths=paths)
        if not data_files:
            console.print("[red]No valid files found to load.")
            raise typer.Exit(1)
    except FileNotValidError as exc:
        console.print(f"[red]{exc.message}")
        raise typer.Exit(1) from exc

    for data_file in data_files:
        if data_file.valid and data_file.content:
            continue
        console.print(f"[red]{data_file.error_message} ({data_file.location})")
        has_error = True

    if has_error:
        raise typer.Exit(1)

    return sorted(data_files, key=lambda x: x.location)


def display_object_validate_format_success(file: ObjectFile, console: Console) -> None:
    if file.multiple_documents:
        console.print(f"[green] File '{file.location}' [{file.document_position}] is Valid!")
    else:
        console.print(f"[green] File '{file.location}' is Valid!")


def display_object_validate_format_error(file: ObjectFile, error: ValidationError, console: Console) -> None:
    if file.multiple_documents:
        console.print(f"[red] File '{file.location}' [{file.document_position}] is not valid!")
    else:
        console.print(f"[red] File '{file.location}' is not valid!")
    if error.messages:
        for message in error.messages:
            console.print(f"[red] {message}")
    else:
        console.print(f"[red] {error.message}")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core import registry
from infrahub.core.initialization import initialization
from infrahub.core.migrations.shared import GraphMigration, MigrationResult
from infrahub.lock import initialize_lock
from infrahub.log import get_logger

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class BackfillMissingHierarchyQuery(Query):
    name = "backfill_missing_hierarchy"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        # load schemas from database into registry
        initialize_lock()
        await initialization(db=db)
        kind_hierarchy_map: dict[str, str] = {}
        schema_branch = await registry.schema.load_schema_from_db(db=db)
        for node_schema_kind in schema_branch.node_names:
            node_schema = schema_branch.get_node(name=node_schema_kind, duplicate=False)
            if node_schema.hierarchy:
                kind_hierarchy_map[node_schema.kind] = node_schema.hierarchy

        self.params = {"hierarchy_map": kind_hierarchy_map}
        query = """
        MATCH (r:Root)
        WITH r.default_branch AS default_branch
        MATCH (rel:Relationship {name: "parent__child"})-[e:IS_RELATED]-(n:Node)
        WHERE e.hierarchy IS NULL
        WITH DISTINCT rel, n, default_branch
        CALL (rel, n, default_branch) {
            MATCH (rel)-[e:IS_RELATED {branch: default_branch}]-(n)
            RETURN e
            ORDER BY e.from DESC
            LIMIT 1
        }
        WITH rel, n, e
        WHERE e.status = "active" AND e.hierarchy IS NULL
        SET e.hierarchy = $hierarchy_map[n.kind]
        """
        self.add_to_query(query)


class Migration024(GraphMigration):
    """
    A bug in diff merge logic caused the hierarchy information on IS_RELATED edges to be lost when merged into
    main. This migration backfills the missing hierarchy data and accounts for the case when the branch that
    created the data has been deleted.
    """

    name: str = "024_backfill_hierarchy"
    minimum_version: int = 23
    queries: Sequence[type[Query]] = [BackfillMissingHierarchyQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.migrations.shared import GraphMigration, MigrationResult
from infrahub.log import get_logger

from ...query import Query, QueryType

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class SetMissingHierarchyQuery(Query):
    name = "set_missing_hierarchy"
    type = QueryType.WRITE
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
        MATCH (r:Root)
        WITH r.default_branch AS default_branch
        MATCH (n:Node)-[main_e:IS_RELATED {branch: default_branch}]-(rel:Relationship)
        WHERE main_e.hierarchy IS NULL
        CALL (n, main_e, rel) {
            MATCH (n)-[branch_e:IS_RELATED]-(rel)
            WHERE branch_e.hierarchy IS NOT NULL
            AND branch_e.branch <> main_e.branch
            AND branch_e.from < main_e.from
            SET main_e.hierarchy = branch_e.hierarchy
        }
        """
        self.add_to_query(query)


class Migration021(GraphMigration):
    """
    A bug in diff merge logic caused the hierarchy information on IS_RELATED edges to be lost when merged into
    main. This migration sets the missing hierarchy data.
    """

    name: str = "021_replace_hierarchy"
    minimum_version: int = 20
    queries: Sequence[type[Query]] = [SetMissingHierarchyQuery]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Field, Int, List, NonNull, ObjectType, String
from infrahub_sdk.utils import extract_fields_first_node
from prefect.client.schemas.objects import StateType

from infrahub.graphql.types.task import TaskNodes, TaskState
from infrahub.task_manager.task import PrefectTask
from infrahub.workflows.constants import WorkflowTag

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class Tasks(ObjectType):
    edges = List(NonNull(TaskNodes), required=True)
    count = Int(required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        limit: int = 10,
        offset: int = 0,
        ids: list[str] | None = None,
        branch: str | None = None,
        state: list | None = None,
        workflow: list[str] | None = None,
        related_node__ids: list | None = None,
        q: str | None = None,
        log_limit: int | None = None,
        log_offset: int | None = None,
    ) -> dict[str, Any]:
        related_nodes = related_node__ids or []
        ids = ids or []
        return await Tasks.query(
            info=info,
            branch=branch,
            limit=limit,
            offset=offset,
            q=q,
            ids=ids,
            statuses=state,
            workflows=workflow,
            related_nodes=related_nodes,
            log_limit=log_limit,
            log_offset=log_offset,
        )

    @staticmethod
    async def resolve_branch_status(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        branch: str,
    ) -> dict[str, Any]:
        statuses: list[StateType] = [StateType.PENDING, StateType.RUNNING, StateType.CANCELLING, StateType.SCHEDULED]
        tags: list[str] = [WorkflowTag.DATABASE_CHANGE.render()]

        return await Tasks.query(info=info, branch=branch, statuses=statuses, tags=tags)

    @classmethod
    async def query(
        cls,
        info: GraphQLResolveInfo,
        related_nodes: list[str] | None = None,
        q: str | None = None,
        ids: list[str] | None = None,
        statuses: list[StateType] | None = None,
        workflows: list[str] | None = None,
        tags: list[str] | None = None,
        branch: str | None = None,
        limit: int | None = None,
        offset: int | None = None,
        log_limit: int | None = None,
        log_offset: int | None = None,
    ) -> dict[str, Any]:
        graphql_context: GraphqlContext = info.context
        fields = await extract_fields_first_node(info)

        prefect_tasks = await PrefectTask.query(
            db=graphql_context.db,
            fields=fields,
            q=q,
            ids=ids,
            branch=branch,
            statuses=statuses,
            workflows=workflows,
            tags=tags,
            related_nodes=related_nodes,
            limit=limit,
            offset=offset,
            log_limit=log_limit,
            log_offset=log_offset,
        )
        prefect_count = prefect_tasks.get("count", None)
        return {
            "count": prefect_count or 0,
            "edges": prefect_tasks.get("edges", []),
        }


Task = Field(
    Tasks,
    limit=Int(required=False),
    offset=Int(required=False),
    related_node__ids=List(String),
    branch=String(required=False),
    state=List(TaskState),
    workflow=List(String),
    ids=List(String),
    q=String(required=False),
    log_limit=Int(required=False),
    log_offset=Int(required=False),
    resolver=Tasks.resolve,
    required=True,
)

TaskBranchStatus = Field(
    Tasks,
    branch=String(required=False),
    description="Return the list of all pending or running tasks that can modify the data, for a given branch",
    resolver=Tasks.resolve_branch_status,
    required=True,
)



import uuid

from pydantic import BaseModel, Field


class PasswordCredential(BaseModel):
    username: str = Field(..., description="Name of the user that is logging in.")
    password: str = Field(..., description="The password of the user.")


class UserToken(BaseModel):
    access_token: str = Field(..., description="JWT access_token")
    refresh_token: str = Field(..., description="JWT refresh_token")


class UserTokenWithUrl(UserToken):
    final_url: str = Field(..., description="The final url after logged in")


class AccessTokenResponse(BaseModel):
    access_token: str = Field(..., description="JWT access_token")


class RefreshTokenData(BaseModel):
    account_id: str
    session_id: uuid.UUID



from __future__ import annotations

import asyncio
import random
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Coroutine, TypeVar

from neo4j import (
    READ_ACCESS,
    WRITE_ACCESS,
    AsyncDriver,
    AsyncGraphDatabase,
    AsyncResult,
    AsyncSession,
    AsyncTransaction,
    NotificationDisabledCategory,
    NotificationMinimumSeverity,
    Query,
    Record,
    TrustAll,
    TrustCustomCAs,
    TrustSystemCAs,
)
from neo4j.exceptions import ClientError, Neo4jError, ServiceUnavailable, TransientError
from opentelemetry import trace
from typing_extensions import Self

from infrahub import config, lock
from infrahub.constants.database import DatabaseType, Neo4jRuntime
from infrahub.core import registry
from infrahub.core.query import QueryType
from infrahub.exceptions import DatabaseError
from infrahub.log import get_logger
from infrahub.utils import InfrahubStringEnum

from .metrics import CONNECTION_POOL_USAGE, QUERY_EXECUTION_METRICS, TRANSACTION_RETRIES

if TYPE_CHECKING:
    from types import TracebackType

    from infrahub.core.branch import Branch
    from infrahub.core.schema import GenericSchema, MainSchemaTypes, NodeSchema
    from infrahub.core.schema.schema_branch import SchemaBranch

validated_database = {}
R = TypeVar("R")

log = get_logger()


@dataclass
class QueryConfig:
    neo4j_runtime: Neo4jRuntime = Neo4jRuntime.DEFAULT
    profile_memory: bool = False


class InfrahubDatabaseMode(InfrahubStringEnum):
    DRIVER = "driver"
    SESSION = "session"
    TRANSACTION = "transaction"


class InfrahubDatabaseSessionMode(InfrahubStringEnum):
    READ = "read"
    WRITE = "write"


def get_branch_name(branch: Branch | str | None = None) -> str:
    if not branch:
        return registry.default_branch
    if isinstance(branch, str):
        return branch

    return branch.name


class DatabaseSchemaManager:
    def __init__(self, db: InfrahubDatabase) -> None:
        self._db = db

    def get(self, name: str, branch: Branch | str | None = None, duplicate: bool = True) -> MainSchemaTypes:
        branch_name = get_branch_name(branch=branch)
        if branch_name not in self._db._schemas:
            return registry.schema.get(name=name, branch=branch, duplicate=duplicate)
        return self._db._schemas[branch_name].get(name=name, duplicate=duplicate)

    def get_node_schema(self, name: str, branch: Branch | str | None = None, duplicate: bool = True) -> NodeSchema:
        schema = self.get(name=name, branch=branch, duplicate=duplicate)
        if schema.is_node_schema:
            return schema

        raise ValueError("The selected node is not of type NodeSchema")

    def get_generic_schema(
        self, name: str, branch: Branch | str | None = None, duplicate: bool = True
    ) -> GenericSchema:
        schema = self.get(name=name, branch=branch, duplicate=duplicate)
        if schema.is_generic_schema:
            return schema

        raise ValueError("The selected node is not of type GenericSchema")

    def set(self, name: str, schema: MainSchemaTypes, branch: str | None = None) -> int:
        branch_name = get_branch_name(branch=branch)
        if branch_name not in self._db._schemas:
            return registry.schema.set(name=name, schema=schema, branch=branch)
        return self._db._schemas[branch_name].set(name=name, schema=schema)

    def has(self, name: str, branch: Branch | str | None = None) -> bool:
        branch_name = get_branch_name(branch=branch)
        if branch_name not in self._db._schemas:
            return registry.schema.has(name=name, branch=branch)
        return self._db._schemas[branch_name].has(name=name)

    def get_full(self, branch: Branch | str | None = None, duplicate: bool = True) -> dict[str, MainSchemaTypes]:
        branch_name = get_branch_name(branch=branch)
        if branch_name not in self._db._schemas:
            return registry.schema.get_full(branch=branch)
        return self._db._schemas[branch_name].get_all(duplicate=duplicate)

    async def get_full_safe(
        self, branch: Branch | str | None = None, duplicate: bool = True
    ) -> dict[str, MainSchemaTypes]:
        await lock.registry.local_schema_wait()
        return self.get_full(branch=branch, duplicate=duplicate)

    def get_schema_branch(self, name: str) -> SchemaBranch:
        if name not in self._db._schemas:
            return registry.schema.get_schema_branch(name=name)
        return self._db._schemas[name]


class InfrahubDatabase:
    """Base class for database access"""

    def __init__(
        self,
        driver: AsyncDriver,
        mode: InfrahubDatabaseMode = InfrahubDatabaseMode.DRIVER,
        db_type: DatabaseType | None = None,
        default_neo4j_runtime: Neo4jRuntime = Neo4jRuntime.DEFAULT,
        schemas: list[SchemaBranch] | None = None,
        session: AsyncSession | None = None,
        session_mode: InfrahubDatabaseSessionMode = InfrahubDatabaseSessionMode.WRITE,
        transaction: AsyncTransaction | None = None,
        queries_names_to_config: dict[str, QueryConfig] | None = None,
    ):
        self._mode: InfrahubDatabaseMode = mode
        self._driver: AsyncDriver = driver
        self._session: AsyncSession | None = session
        self._session_mode: InfrahubDatabaseSessionMode = session_mode
        self._is_session_local: bool = False
        self._transaction: AsyncTransaction | None = transaction
        self.default_neo4j_runtime = default_neo4j_runtime
        self.queries_names_to_config = queries_names_to_config if queries_names_to_config is not None else {}

        if schemas:
            self._schemas: dict[str, SchemaBranch] = {schema.name: schema for schema in schemas}
        else:
            self._schemas = {}
        self.schema: DatabaseSchemaManager = DatabaseSchemaManager(db=self)

        if db_type and isinstance(db_type, DatabaseType):
            self.db_type = db_type
        else:
            self.db_type = config.SETTINGS.database.db_type

    def __del__(self) -> None:
        if not self._session or not self._is_session_local or self._session.closed():
            return

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None
        if loop and loop.is_running():
            loop.create_task(self._session.close())
        else:
            asyncio.run(self._session.close())

    @property
    def is_session(self) -> bool:
        if self._mode == InfrahubDatabaseMode.SESSION:
            return True
        return False

    @property
    def is_transaction(self) -> bool:
        if self._mode == InfrahubDatabaseMode.TRANSACTION:
            return True
        return False

    def get_context(self) -> dict[str, Any]:
        """
        This method is meant to be overridden by subclasses in order to fill in subclass attributes
        to methods returning a copy of this object using self.__class__ constructor.
        """

        return {}

    def add_schema(self, schema: SchemaBranch, name: str | None = None) -> None:
        self._schemas[name or schema.name] = schema

    def purge_inactive_schemas(self, active_branches: list[str]) -> list[str]:
        """Return non active schema branches that were purged."""
        removed_branches: list[str] = []
        for branch_name in list(self._schemas.keys()):
            if branch_name not in active_branches:
                del self._schemas[branch_name]
                removed_branches.append(branch_name)

        return removed_branches

    def start_session(self, read_only: bool = False, schemas: list[SchemaBranch] | None = None) -> InfrahubDatabase:
        """Create a new InfrahubDatabase object in Session mode."""
        session_mode = InfrahubDatabaseSessionMode.WRITE
        if read_only:
            session_mode = InfrahubDatabaseSessionMode.READ

        context = self.get_context()

        return self.__class__(
            mode=InfrahubDatabaseMode.SESSION,
            db_type=self.db_type,
            default_neo4j_runtime=self.default_neo4j_runtime,
            schemas=schemas or self._schemas.values(),
            driver=self._driver,
            session_mode=session_mode,
            queries_names_to_config=self.queries_names_to_config,
            **context,
        )

    def start_transaction(self, schemas: list[SchemaBranch] | None = None) -> InfrahubDatabase:
        context = self.get_context()

        return self.__class__(
            mode=InfrahubDatabaseMode.TRANSACTION,
            db_type=self.db_type,
            default_neo4j_runtime=self.default_neo4j_runtime,
            schemas=schemas or self._schemas.values(),
            driver=self._driver,
            session=self._session,
            session_mode=self._session_mode,
            queries_names_to_config=self.queries_names_to_config,
            **context,
        )

    async def session(self) -> AsyncSession:
        if self._session:
            return self._session

        if self._session_mode == InfrahubDatabaseSessionMode.READ:
            self._session = self._driver.session(
                database=config.SETTINGS.database.database_name, default_access_mode=READ_ACCESS
            )
        else:
            self._session = self._driver.session(
                database=config.SETTINGS.database.database_name, default_access_mode=WRITE_ACCESS
            )

        self._is_session_local = True
        return self._session

    async def transaction(self, name: str | None) -> AsyncTransaction:
        if self._transaction:
            return self._transaction

        session = await self.session()
        self._transaction = await session.begin_transaction(
            metadata={"name": name, "infrahub_id": f"{trace.get_current_span().get_span_context().span_id:x}"}
        )
        return self._transaction

    async def __aenter__(self) -> Self:
        if self._mode == InfrahubDatabaseMode.SESSION:
            if self._session_mode == InfrahubDatabaseSessionMode.READ:
                self._session = self._driver.session(
                    database=config.SETTINGS.database.database_name, default_access_mode=READ_ACCESS
                )
            else:
                self._session = self._driver.session(
                    database=config.SETTINGS.database.database_name, default_access_mode=WRITE_ACCESS
                )

        elif self._mode == InfrahubDatabaseMode.TRANSACTION:
            session = await self.session()
            self._transaction = await session.begin_transaction()

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if self._mode == InfrahubDatabaseMode.SESSION:
            return await self._session.close()

        if self._mode == InfrahubDatabaseMode.TRANSACTION:
            if exc_type is not None:
                await self._transaction.rollback()
            else:
                try:
                    await self._transaction.commit()
                except Neo4jError as exc:
                    raise exc
                finally:
                    await self._transaction.close()

            if self._is_session_local:
                await self._session.close()

    async def close(self) -> None:
        await self._driver.close()

    async def execute_query(
        self,
        query: str,
        params: dict[str, Any] | None = None,
        name: str = "undefined",
        context: dict[str, str] | None = None,
        type: QueryType | None = None,
    ) -> list[Record]:
        results, _ = await self.execute_query_with_metadata(
            query=query, params=params, name=name, context=context, type=type
        )
        return results

    async def execute_query_with_metadata(
        self,
        query: str,
        params: dict[str, Any] | None = None,
        name: str = "undefined",
        context: dict[str, str] | None = None,
        type: QueryType | None = None,
    ) -> tuple[list[Record], dict[str, Any]]:
        connpool_usage = self._driver._pool.in_use_connection_count(self._driver._pool.address)
        CONNECTION_POOL_USAGE.labels(self._driver._pool.address).set(float(connpool_usage))

        if config.SETTINGS.database.max_concurrent_queries:
            while connpool_usage > config.SETTINGS.database.max_concurrent_queries:
                await asyncio.sleep(config.SETTINGS.database.max_concurrent_queries_delay)
                connpool_usage = self._driver._pool.in_use_connection_count(self._driver._pool.address)

        with trace.get_tracer(__name__).start_as_current_span("execute_db_query_with_metadata") as span:
            span.set_attribute("query", query)
            if name:
                span.set_attribute("query_name", name)

            runtime = self.default_neo4j_runtime

            if self.db_type == DatabaseType.NEO4J:
                query_config = self.queries_names_to_config.get(name, None)

                if query_config and query_config.neo4j_runtime not in [Neo4jRuntime.DEFAULT, Neo4jRuntime.UNDEFINED]:
                    runtime = query_config.neo4j_runtime

                if (
                    type
                    and type == QueryType.READ
                    and runtime not in [Neo4jRuntime.DEFAULT, Neo4jRuntime.UNDEFINED]
                    and not (self.is_transaction and runtime in [Neo4jRuntime.PARALLEL])
                ):
                    query = f"CYPHER runtime = {runtime.value}\n" + query
                else:
                    runtime = Neo4jRuntime.DEFAULT

                if query_config and query_config.profile_memory:
                    query = "PROFILE\n" + query

            labels = {
                "type": type.value if type else self._session_mode.value,
                "query": name,
                "runtime": runtime.value,
                "context1": "",
                "context2": "",
            }
            if context:
                labels.update(
                    {
                        f"context{idx + 1}": f"{key}__{value}"
                        for idx, (key, value) in enumerate(context.items())
                        if idx <= 1
                    }
                )

            with QUERY_EXECUTION_METRICS.labels(**labels).time():
                response = await self.run_query(query=query, params=params, name=name)
                if response is None:
                    span.set_attribute("rows", "empty")
                    return [], {}
                results = [item async for item in response]
                span.set_attribute("rows", len(results))
                return results, response._metadata or {}

    async def run_query(
        self, query: str, params: dict[str, Any] | None = None, name: str | None = "undefined"
    ) -> AsyncResult:
        _query: str | Query = query
        if self.is_transaction:
            execution_method = await self.transaction(name=name)
        else:
            _query = Query(
                text=query,
                metadata={"name": name, "infrahub_id": f"{trace.get_current_span().get_span_context().span_id:x}"},
            )
            execution_method = await self.session()

        try:
            response = await execution_method.run(query=_query, parameters=params)
        except ServiceUnavailable as exc:
            log.error("Database Service unavailable", error=str(exc))
            raise DatabaseError(message="Unable to connect to the database") from exc

        return response

    def render_list_comprehension(self, items: str, item_name: str) -> str:
        if self.db_type == DatabaseType.MEMGRAPH:
            return f"extract(i in {items} | i.{item_name})"
        return f"[i IN {items} | i.{item_name}]"

    def render_list_comprehension_with_list(self, items: str, item_names: list[str]) -> str:
        item_names_str = ",".join([f"i.{name}" for name in item_names])
        if self.db_type == DatabaseType.MEMGRAPH:
            return f"extract(i in {items} | [{item_names_str}])"
        return f"[i IN {items} | [{item_names_str}]]"

    def render_uuid_generation(self, node_label: str, node_attr: str, index: int = 1) -> str:
        generate_uuid_query = f"SET {node_label}.{node_attr} = randomUUID()"
        if self.db_type == DatabaseType.MEMGRAPH:
            generate_uuid_query = f"""
            CALL uuid_generator.get() YIELD uuid AS uuid{index}
            SET {node_label}.{node_attr} = uuid{index}
            """
        return generate_uuid_query

    def get_id_function_name(self) -> str:
        if self.db_type == DatabaseType.NEO4J:
            return "elementId"
        return "ID"

    def to_database_id(self, db_id: str | int) -> str | int:
        if self.db_type == DatabaseType.NEO4J:
            return db_id
        try:
            return int(db_id)
        except ValueError:
            return db_id


async def create_database(driver: AsyncDriver, database_name: str) -> None:
    default_db = driver.session()
    await default_db.run(f"CREATE DATABASE {database_name} WAIT")


async def validate_database(
    driver: AsyncDriver, database_name: str, retry: int = 0, retry_interval: int = 1, create_db: bool = True
) -> bool:
    """Validate if a database is present in Neo4j by executing a simple query.

    Args:
        driver (AsyncDriver): Neo4j Driver
        database_name (str): Name of the database in Neo4j
        retry (int, optional): Number of retry before raising an exception. Defaults to 0.
        retry_interval (int, optional): Time between retries in second. Defaults to 1.
    """

    try:
        session = driver.session(database=database_name)
        await session.run("SHOW TRANSACTIONS")
        validated_database[database_name] = True

    except ClientError as exc:
        if create_db and exc.code == "Neo.ClientError.Database.DatabaseNotFound":
            await create_database(driver=driver, database_name=config.SETTINGS.database.database_name)

        if retry == 0:
            raise

        await asyncio.sleep(retry_interval)
        await validate_database(driver=driver, database_name=database_name, retry=retry - 1, create_db=False)

    return True


async def get_db(retry: int = 0) -> AsyncDriver:
    trusted_certificates = TrustSystemCAs()
    if config.SETTINGS.database.tls_insecure:
        trusted_certificates = TrustAll()
    elif config.SETTINGS.database.tls_ca_file:
        trusted_certificates = TrustCustomCAs(config.SETTINGS.database.tls_ca_file)

    driver = AsyncGraphDatabase.driver(
        config.SETTINGS.database.database_uri,
        auth=(config.SETTINGS.database.username, config.SETTINGS.database.password),
        encrypted=config.SETTINGS.database.tls_enabled,
        trusted_certificates=trusted_certificates,
        notifications_disabled_categories=[
            NotificationDisabledCategory.UNRECOGNIZED,
        ],
        notifications_min_severity=NotificationMinimumSeverity.WARNING,
    )

    if config.SETTINGS.database.database_name not in validated_database:
        await validate_database(
            driver=driver, database_name=config.SETTINGS.database.database_name, retry=retry, create_db=True
        )

    return driver


def retry_db_transaction(
    name: str,
) -> Callable[[Callable[..., Coroutine[Any, Any, R]]], Callable[..., Coroutine[Any, Any, R]]]:
    def func_wrapper(func: Callable[..., Coroutine[Any, Any, R]]) -> Callable[..., Coroutine[Any, Any, R]]:
        async def wrapper(*args: Any, **kwargs: Any) -> R:
            error = Exception()
            for attempt in range(1, config.SETTINGS.database.retry_limit + 1):
                try:
                    return await func(*args, **kwargs)
                except (TransientError, ClientError) as exc:
                    if isinstance(exc, ClientError):
                        if exc.code != "Neo.ClientError.Statement.EntityNotFound":
                            raise exc
                    retry_time: float = random.randrange(100, 500) / 1000
                    log.exception("Retry handler caught database error")
                    log.info(
                        f"Retrying database transaction, attempt {attempt}/{config.SETTINGS.database.retry_limit}",
                        retry_time=retry_time,
                    )
                    log.debug("Database transaction failed", message=exc.message)
                    TRANSACTION_RETRIES.labels(name).inc()
                    await asyncio.sleep(retry_time)
                    if attempt == config.SETTINGS.database.retry_limit:
                        error = exc
                        break
            raise error

        return wrapper

    return func_wrapper



from infrahub.core.manager import NodeManager
from infrahub.core.protocols import CoreMenuItem
from infrahub.database import InfrahubDatabase

from .models import MenuDict, MenuItemDefinition, MenuItemDict


class MenuRepository:
    def __init__(self, db: InfrahubDatabase):
        self.db = db

    async def get_menu(self, nodes: dict[str, CoreMenuItem] | None = None) -> MenuDict:
        menu_nodes = nodes or await self.get_menu_db()
        return await self._convert_menu_from_db(nodes=menu_nodes)

    async def _convert_menu_from_db(self, nodes: dict[str, CoreMenuItem]) -> MenuDict:
        menu = MenuDict()
        menu_by_ids = {menu_node.get_id(): MenuItemDict.from_node(menu_node) for menu_node in nodes.values()}

        async def add_children(menu_item: MenuItemDict, menu_node: CoreMenuItem) -> MenuItemDict:
            children = await menu_node.children.get_peers(db=self.db, peer_type=CoreMenuItem)
            for child_id, child_node in children.items():
                if child_menu_item := menu_by_ids.get(child_id):
                    child = await add_children(child_menu_item, child_node)
                    menu_item.children[str(child.identifier)] = child
            return menu_item

        for menu_node in nodes.values():
            menu_item = menu_by_ids[menu_node.get_id()]
            parent = await menu_node.parent.get_peer(db=self.db, peer_type=CoreMenuItem)
            if parent:
                continue

            children = await menu_node.children.get_peers(db=self.db, peer_type=CoreMenuItem)
            for child_id, child_node in children.items():
                if child_menu_item := menu_by_ids.get(child_id):
                    child = await add_children(child_menu_item, child_node)
                    menu_item.children[str(child.identifier)] = child

            menu.data[str(menu_item.identifier)] = menu_item

        return menu

    async def get_menu_db(self) -> dict[str, CoreMenuItem]:
        menu_nodes = await NodeManager.query(
            schema=CoreMenuItem,
            filters={"namespace__value": "Builtin"},
            prefetch_relationships=True,
            db=self.db,
        )
        return {node.get_id(): node for node in menu_nodes}

    async def create_menu(self, menu: list[MenuItemDefinition]) -> None:
        for item in menu:
            obj = await item.to_node(db=self.db)
            await obj.save(db=self.db)
            if item.children:
                await self.create_menu_children(parent=obj, children=item.children)

    async def create_menu_children(self, parent: CoreMenuItem, children: list[MenuItemDefinition]) -> None:
        for child in children:
            obj = await child.to_node(db=self.db, parent=parent)
            await obj.save(db=self.db)
            if child.children:
                await self.create_menu_children(parent=obj, children=child.children)

    async def update_menu(
        self, existing_menu: MenuDict, new_menu: MenuDict, menu_nodes: dict[str, CoreMenuItem]
    ) -> None:
        async def process_menu_item(menu_item: MenuItemDict, parent: CoreMenuItem | None) -> None:
            existing_item = existing_menu.find_item(name=str(menu_item.identifier))
            if existing_item and existing_item.id:
                node = menu_nodes[existing_item.id]
                await self.update_menu_item(
                    node=node, existing_menu_item=existing_item, new_menu_item=menu_item, parent=parent
                )
            else:
                node = await self.create_menu_item(new_menu_item=menu_item, parent=parent)

            for child_item in menu_item.children.values():
                await process_menu_item(menu_item=child_item, parent=node)

        for top_level_item in new_menu.data.values():
            await process_menu_item(menu_item=top_level_item, parent=None)

        # Delete items that are not in the new menu
        menu_to_delete = existing_menu.get_all_identifiers() - new_menu.get_all_identifiers()
        for item_to_delete in menu_to_delete:
            existing_item = existing_menu.find_item(name=item_to_delete)
            if existing_item and existing_item.id:
                node = menu_nodes[existing_item.id]
                await node.delete(db=self.db)

    async def update_menu_item(
        self,
        node: CoreMenuItem,
        existing_menu_item: MenuItemDict,
        new_menu_item: MenuItemDict,
        parent: CoreMenuItem | None,
    ) -> None:
        attrs_to_update = existing_menu_item.diff_attributes(new_menu_item)
        for attr_name, value in attrs_to_update.items():
            attr = getattr(node, attr_name)
            attr.value = value
        await node.parent.update(data=parent, db=self.db)  # type: ignore[arg-type]
        await node.save(db=self.db)

    async def create_menu_item(self, new_menu_item: MenuItemDict, parent: CoreMenuItem | None) -> CoreMenuItem:
        obj = await new_menu_item.to_node(db=self.db, parent=parent)
        await obj.save(db=self.db)
        return obj



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub import config
from infrahub.core.manager import NodeManager
from infrahub.core.registry import registry
from infrahub.exceptions import SchemaNotFoundError
from infrahub.log import get_logger

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


log = get_logger(__name__)


async def get_display_labels_per_kind(
    kind: str, ids: list[str], branch_name: str, db: InfrahubDatabase, skip_missing_schema: bool = False
) -> dict[str, str]:
    """Return the display_labels of a list of nodes of a specific kind."""
    branch = await registry.get_branch(branch=branch_name, db=db)
    schema_branch = db.schema.get_schema_branch(name=branch.name)
    try:
        fields = schema_branch.generate_fields_for_display_label(name=kind)
    except SchemaNotFoundError:
        if skip_missing_schema:
            return {}
        raise
    display_label_map: dict[str, str] = {}
    offset = 0
    limit = config.SETTINGS.database.query_size_limit
    while True:
        limited_ids = ids[offset : offset + limit]
        if not limited_ids:
            break
        node_map = await NodeManager.get_many(ids=limited_ids, fields=fields, db=db, branch=branch)
        for node_id, node in node_map.items():
            display_label_map[node_id] = await node.render_display_label(db=db)
        offset += limit
    return display_label_map


async def get_display_labels(nodes: dict[str, dict[str, list[str]]], db: InfrahubDatabase) -> dict[str, dict[str, str]]:
    """Query the display_labels of a group of nodes organized per branch and per kind."""
    response: dict[str, dict[str, str]] = {}
    for branch_name, items in nodes.items():
        if branch_name not in response:
            response[branch_name] = {}
        for kind, ids in items.items():
            labels = await get_display_labels_per_kind(
                kind=kind, ids=ids, db=db, branch_name=branch_name, skip_missing_schema=True
            )
            response[branch_name].update(labels)

    return response



import ipaddress
from typing import TYPE_CHECKING

from infrahub.core.branch import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.query.ipam import IPPrefixReconcileQuery
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase, retry_db_transaction
from infrahub.exceptions import NodeNotFoundError

from .constants import AllIPTypes

if TYPE_CHECKING:
    from infrahub.core.relationship.model import RelationshipManager


class IPNodesToReconcile:
    def __init__(
        self,
        node_uuid: str,
        current_parent_uuid: str | None,
        calculated_parent_uuid: str | None,
        current_child_uuids: set[str],
        calculated_child_uuids: set[str],
        node_map: dict[str, Node],
    ) -> None:
        self.node_uuid = node_uuid
        self.current_parent_uuid = current_parent_uuid
        self.calculated_parent_uuid = calculated_parent_uuid
        self.current_child_uuids = current_child_uuids
        self.calculated_child_uuids = calculated_child_uuids
        self.node_map = node_map
        self._calculated_child_nodes = set()
        self._current_child_nodes = set()
        for ccu in self.current_child_uuids:
            if ccu in self.node_map:
                self._current_child_nodes.add(self.node_map[ccu])
        for ccu in self.calculated_child_uuids:
            if ccu in self.node_map:
                self._calculated_child_nodes.add(self.node_map[ccu])

    @property
    def node(self) -> Node:
        return self.node_map[self.node_uuid]

    @property
    def current_parent(self) -> Node | None:
        if not self.current_parent_uuid:
            return None
        return self.node_map.get(self.current_parent_uuid)

    @property
    def calculated_parent(self) -> Node | None:
        if not self.calculated_parent_uuid:
            return None
        return self.node_map.get(self.calculated_parent_uuid)

    @property
    def current_child_nodes(self) -> set[Node]:
        return self._current_child_nodes

    @property
    def calculated_child_nodes(self) -> set[Node]:
        return self._calculated_child_nodes

    def get_node_by_uuid(self, uuid: str) -> Node:
        return self.node_map[uuid]

    async def _get_child_uuids(self, db: InfrahubDatabase, node: Node | None) -> set[str]:
        if not node:
            return set()
        child_uuids = set()
        child_prefix_rels = await node.children.get_relationships(db=db, force_refresh=False)  # type: ignore[attr-defined]
        child_uuids |= {cpr.get_peer_id() for cpr in child_prefix_rels}
        child_address_rels = await node.ip_addresses.get_relationships(db=db, force_refresh=False)  # type: ignore[attr-defined]
        child_uuids |= {car.get_peer_id() for car in child_address_rels}
        return child_uuids


class IpamReconciler:
    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self.db = db
        self.branch = branch
        self.at: Timestamp | None = None

    @retry_db_transaction(name="ipam_reconcile")
    async def reconcile(
        self,
        ip_value: AllIPTypes,
        namespace: Node | str | None = None,
        node_uuid: str | None = None,
        is_delete: bool = False,
        at: Timestamp | None = None,
    ) -> Node | None:
        self.at = Timestamp(at)

        query = await IPPrefixReconcileQuery.init(
            db=self.db, branch=self.branch, ip_value=ip_value, namespace=namespace, node_uuid=node_uuid, at=self.at
        )
        await query.execute(db=self.db)

        ip_node_uuid = query.get_ip_node_uuid()
        if not ip_node_uuid:
            node_type = InfrahubKind.IPPREFIX
            if isinstance(ip_value, ipaddress.IPv6Interface | ipaddress.IPv4Interface):
                node_type = InfrahubKind.IPADDRESS
            raise NodeNotFoundError(node_type=node_type, identifier=str(ip_value))
        current_parent_uuid = query.get_current_parent_uuid()
        calculated_parent_uuid = query.get_calculated_parent_uuid()
        current_children_uuids = set(query.get_current_children_uuids())
        calculated_children_uuids = set(query.get_calculated_children_uuids())

        all_uuids: set[str] = set()
        all_uuids = (all_uuids | {ip_node_uuid}) if ip_node_uuid else all_uuids
        all_uuids = (all_uuids | {current_parent_uuid}) if current_parent_uuid else all_uuids
        all_uuids = (all_uuids | {calculated_parent_uuid}) if calculated_parent_uuid else all_uuids
        all_uuids |= current_children_uuids
        all_uuids |= calculated_children_uuids
        all_nodes = await NodeManager.get_many(
            db=self.db,
            branch=self.branch,
            ids=list(all_uuids),
        )

        reconcile_nodes = IPNodesToReconcile(
            node_uuid=ip_node_uuid,
            current_parent_uuid=current_parent_uuid,
            calculated_parent_uuid=calculated_parent_uuid,
            current_child_uuids=current_children_uuids,
            calculated_child_uuids=calculated_children_uuids,
            node_map=all_nodes,
        )

        if is_delete:
            updated_uuids = await self.update_children_for_delete(reconcile_nodes)
        else:
            updated_uuids = await self.update_node(reconcile_nodes)
            updated_uuids |= await self.update_current_children(reconcile_nodes)
            updated_uuids |= await self.update_calculated_children(reconcile_nodes)

        for updated_uuid in updated_uuids:
            node = reconcile_nodes.get_node_by_uuid(updated_uuid)
            await node.save(db=self.db, at=self.at)

        if is_delete:
            try:
                await reconcile_nodes.node.delete(db=self.db, at=self.at)
            except KeyError:
                return None

        return reconcile_nodes.node

    async def _update_node_parent(self, node: Node, new_parent_uuid: str | None) -> None:
        node_kinds = {node.get_kind()} | set(node.get_schema().inherit_from)
        is_prefix = False
        if InfrahubKind.IPADDRESS in node_kinds:
            rel_manager: RelationshipManager = node.ip_prefix  # type: ignore[attr-defined]
        elif InfrahubKind.IPPREFIX in node_kinds:
            rel_manager = node.parent  # type: ignore[attr-defined]
            is_prefix = True
        else:
            return

        await rel_manager.update(db=self.db, data=new_parent_uuid)
        if not is_prefix:
            return
        node.is_top_level.value = new_parent_uuid is None  # type: ignore[attr-defined]

    async def update_node(self, reconcile_nodes: IPNodesToReconcile) -> set[str]:
        await self._update_node_parent(
            node=reconcile_nodes.node, new_parent_uuid=reconcile_nodes.calculated_parent_uuid
        )
        return {reconcile_nodes.node.get_id()}

    async def update_node_for_delete(self, reconcile_nodes: IPNodesToReconcile) -> set[str]:
        await self._update_node_parent(node=reconcile_nodes.node, new_parent_uuid=None)
        return {reconcile_nodes.node.get_id()}

    async def update_current_children(self, reconcile_nodes: IPNodesToReconcile) -> set[str]:
        updated_uuids = set()
        for current_child_node in reconcile_nodes.current_child_nodes:
            current_child_uuid = current_child_node.get_id()
            if current_child_uuid in reconcile_nodes.calculated_child_uuids:
                # current child is still a child, no update necessary
                continue
            # set parent of deleted child to current_parent of the node (might be None)
            await self._update_node_parent(node=current_child_node, new_parent_uuid=reconcile_nodes.current_parent_uuid)
            updated_uuids.add(current_child_uuid)
        return updated_uuids

    async def update_calculated_children(self, reconcile_nodes: IPNodesToReconcile) -> set[str]:
        updated_uuids = set()
        for calculated_child_node in reconcile_nodes.calculated_child_nodes:
            calculated_child_uuid = calculated_child_node.get_id()
            if calculated_child_uuid in reconcile_nodes.current_child_uuids:
                # calculated child is already a child, no update necessary
                continue
            # set parent of new child to the node
            await self._update_node_parent(node=calculated_child_node, new_parent_uuid=reconcile_nodes.node_uuid)
            updated_uuids.add(calculated_child_uuid)
        return updated_uuids

    async def update_children_for_delete(self, reconcile_nodes: IPNodesToReconcile) -> set[str]:
        updated_uuids = set()
        for current_child_node in reconcile_nodes.current_child_nodes:
            current_child_uuid = current_child_node.get_id()
            await self._update_node_parent(node=current_child_node, new_parent_uuid=reconcile_nodes.current_parent_uuid)
            updated_uuids.add(current_child_uuid)
        return updated_uuids



from infrahub.core.validators.uniqueness.checker import UniquenessChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaUniquenessConstraintDependency(DependencyBuilder[UniquenessChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> UniquenessChecker:
        return UniquenessChecker(db=context.db, branch=context.branch)



from __future__ import annotations

import importlib.metadata

from .client import InfrahubClient, InfrahubClientSync
from .config import Config

__all__ = [
    "Config",
    "InfrahubClient",
    "InfrahubClientSync",
]

try:
    __version__ = importlib.metadata.version("infrahub-sdk")
except importlib.metadata.PackageNotFoundError:
    __version__ = importlib.metadata.version("infrahub-server")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, Field, String

from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME

from .standard_node import InfrahubObjectType

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext


class BranchType(InfrahubObjectType):
    id = String(required=True)
    name = String(required=True)
    description = String(required=False)
    origin_branch = String(required=False)
    branched_from = String(required=False)
    created_at = String(required=False)
    sync_with_git = Boolean(required=False)
    is_default = Boolean(required=False)
    is_isolated = Field(Boolean(required=False), deprecation_reason="non isolated mode is not supported anymore")
    has_schema_changes = Boolean(required=False)

    class Meta:
        description = "Branch"
        name = "Branch"
        model = Branch

    @classmethod
    async def get_list(
        cls,
        fields: dict,
        graphql_context: GraphqlContext,
        **kwargs: Any,
    ) -> list[dict[str, Any]]:
        async with graphql_context.db.start_session(read_only=True) as db:
            objs = await Branch.get_list(db=db, **kwargs)

            if not objs:
                return []

            return [await obj.to_graphql(fields=fields) for obj in objs if obj.name != GLOBAL_BRANCH_NAME]



from pathlib import Path

import yaml

from ..schema.repository import InfrahubRepositoryConfig
from .exceptions import FileNotValidError


def load_repository_config(repo_config_file: Path) -> InfrahubRepositoryConfig:
    if not repo_config_file.is_file():
        raise FileNotFoundError(repo_config_file)

    try:
        yaml_data = repo_config_file.read_text()
        data = yaml.safe_load(yaml_data)
    except yaml.YAMLError as exc:
        raise FileNotValidError(name=str(repo_config_file)) from exc

    return InfrahubRepositoryConfig(**data)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel

from infrahub import config
from infrahub.core.constants import RelationshipDirection
from infrahub.core.query import QueryNode, QueryRel, QueryRelDirection
from infrahub.core.relationship import Relationship
from infrahub.exceptions import InitializationError

from .generated.relationship_schema import GeneratedRelationshipSchema

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.query import QueryElement
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.database import InfrahubDatabase


class RelationshipSchema(GeneratedRelationshipSchema):
    _exclude_from_hash: list[str] = ["filters"]
    _sort_by: list[str] = ["name"]

    @property
    def is_attribute(self) -> bool:
        return False

    @property
    def is_relationship(self) -> bool:
        return True

    @property
    def is_deprecated(self) -> bool:
        return bool(self.deprecation)

    def to_dict(self) -> dict:
        data = self.model_dump(exclude_unset=True, exclude_none=True)
        for field_name, value in data.items():
            if isinstance(value, Enum):
                data[field_name] = value.value
        return data

    def get_class(self) -> type[Relationship]:
        return Relationship

    def get_peer_schema(self, db: InfrahubDatabase, branch: Branch | str | None = None) -> MainSchemaTypes:
        return db.schema.get(name=self.peer, branch=branch, duplicate=False)

    @property
    def internal_peer(self) -> bool:
        return self.peer.startswith("Internal")

    def get_identifier(self) -> str:
        if not self.identifier:
            raise ValueError("RelationshipSchema is not initialized")
        return self.identifier

    def get_id(self) -> str:
        if not self.id:
            raise InitializationError("The relationship schema has not been saved yet and doesn't have an id")
        return self.id

    def get_query_arrows(self) -> QueryArrows:
        """Return (in 4 parts) the 2 arrows for the relationship R1 and R2 based on the direction of the relationship."""

        if self.direction == RelationshipDirection.OUTBOUND:
            return QueryArrows(left=QueryArrowOutband(), right=QueryArrowOutband())
        if self.direction == RelationshipDirection.INBOUND:
            return QueryArrows(left=QueryArrowInband(), right=QueryArrowInband())

        return QueryArrows(left=QueryArrowOutband(), right=QueryArrowInband())

    def update_from_generic(self, other: RelationshipSchema) -> None:
        fields_to_exclude = ("id", "order_weight", "branch", "inherited", "filters")
        for name in self.model_fields:
            if name in fields_to_exclude:
                continue
            if getattr(self, name) != getattr(other, name):
                setattr(self, name, getattr(other, name))

    async def get_query_filter(
        self,
        db: InfrahubDatabase,
        filter_name: str,
        filter_value: str | int | bool | None = None,
        name: str | None = None,  # noqa: ARG002
        branch: Branch | None = None,
        include_match: bool = True,
        param_prefix: str | None = None,
        partial_match: bool = False,
        support_profiles: bool = False,  # noqa: ARG002
    ) -> tuple[list[QueryElement], dict[str, Any], list[str]]:
        """Generate Query String Snippet to filter the right node."""

        query_filter: list[QueryElement] = []
        query_params: dict[str, Any] = {}
        query_where: list[str] = []

        prefix = param_prefix or f"rel_{self.name}"

        query_params[f"{prefix}_rel_name"] = self.identifier

        rel_type = self.get_class().rel_type
        peer_schema = self.get_peer_schema(db=db, branch=branch)

        if include_match:
            query_filter.append(QueryNode(name="n"))

        # Determine in which direction the relationships should point based on the side of the query
        rels_direction = {
            "r1": QueryRelDirection.OUTBOUND,
            "r2": QueryRelDirection.INBOUND,
        }
        if self.direction == RelationshipDirection.OUTBOUND:
            rels_direction = {
                "r1": QueryRelDirection.OUTBOUND,
                "r2": QueryRelDirection.OUTBOUND,
            }
        if self.direction == RelationshipDirection.INBOUND:
            rels_direction = {
                "r1": QueryRelDirection.INBOUND,
                "r2": QueryRelDirection.INBOUND,
            }

        if filter_name == "id":
            query_filter.extend(
                [
                    QueryRel(name="r1", labels=[rel_type], direction=rels_direction["r1"]),
                    QueryNode(name="rl", labels=["Relationship"], params={"name": f"${prefix}_rel_name"}),
                    QueryRel(name="r2", labels=[rel_type], direction=rels_direction["r2"]),
                    QueryNode(name="peer", labels=["Node"]),
                ]
            )

            if filter_value is not None:
                query_filter[-1].params = {"uuid": f"${prefix}_peer_id"}
                query_params[f"{prefix}_peer_id"] = filter_value

            return query_filter, query_params, query_where

        if filter_name in ("ids", "isnull"):
            query_filter.extend(
                [
                    QueryRel(name="r1", labels=[rel_type], direction=rels_direction["r1"]),
                    QueryNode(name="rl", labels=["Relationship"], params={"name": f"${prefix}_rel_name"}),
                    QueryRel(name="r2", labels=[rel_type], direction=rels_direction["r2"]),
                    QueryNode(name="peer", labels=["Node"]),
                ]
            )

            if filter_name == "ids" and filter_value is not None:
                query_params[f"{prefix}_peer_ids"] = filter_value
                query_where.append(f"peer.uuid IN ${prefix}_peer_ids")

            return query_filter, query_params, query_where

        if "__" not in filter_name:
            return query_filter, query_params, query_where

        # -------------------------------------------------------------------
        # Check if the filter is matching
        # -------------------------------------------------------------------
        filter_field_name, filter_next_name = filter_name.split("__", maxsplit=1)

        if filter_field_name not in peer_schema.valid_input_names:
            return query_filter, query_params, query_where

        if self.hierarchical:
            query_filter.extend(
                [
                    QueryRel(
                        labels=[rel_type],
                        direction=rels_direction["r1"],
                        length_min=2,
                        length_max=config.SETTINGS.database.max_depth_search_hierarchy * 2,
                        params={"hierarchy": self.hierarchical},
                    ),
                    QueryNode(name="peer", labels=[self.hierarchical]),
                ]
            )
        else:
            query_filter.extend(
                [
                    QueryRel(name="r1", labels=[rel_type], direction=rels_direction["r1"]),
                    QueryNode(name="rl", labels=["Relationship"], params={"name": f"${prefix}_rel_name"}),
                    QueryRel(name="r2", labels=[rel_type], direction=rels_direction["r2"]),
                    QueryNode(name="peer", labels=["Node"]),
                ]
            )

        field = peer_schema.get_field(filter_field_name)

        field_filter, field_params, field_where = await field.get_query_filter(
            db=db,
            name=filter_field_name,
            filter_name=filter_next_name,
            filter_value=filter_value,
            branch=branch,
            include_match=False,
            param_prefix=prefix if param_prefix else None,
            partial_match=partial_match,
        )

        query_filter.extend(field_filter)
        query_where.extend(field_where)
        query_params.update(field_params)

        return query_filter, query_params, query_where


class QueryArrow(BaseModel):
    start: str
    end: str


class QueryArrowInband(QueryArrow):
    start: str = "<-"
    end: str = "-"


class QueryArrowOutband(QueryArrow):
    start: str = "-"
    end: str = "->"


class QueryArrowBidir(QueryArrow):
    start: str = "-"
    end: str = "-"


class QueryArrows(BaseModel):
    left: QueryArrow
    right: QueryArrow



from __future__ import annotations

import enum
from enum import Enum
from typing import TYPE_CHECKING, Any, Self

from pydantic import Field, ValidationInfo, field_validator, model_validator

from infrahub import config
from infrahub.core.constants.schema import UpdateSupport
from infrahub.core.enums import generate_python_enum
from infrahub.core.query.attribute import default_attribute_query_filter
from infrahub.exceptions import InitializationError
from infrahub.types import ATTRIBUTE_KIND_LABELS, ATTRIBUTE_TYPES

from .attribute_parameters import (
    AttributeParameters,
    NumberAttributeParameters,
    NumberPoolParameters,
    TextAttributeParameters,
    get_attribute_parameters_class_for_kind,
)
from .generated.attribute_schema import GeneratedAttributeSchema

if TYPE_CHECKING:
    from infrahub.core.attribute import BaseAttribute
    from infrahub.core.branch import Branch
    from infrahub.core.constants import BranchSupportType
    from infrahub.core.query import QueryElement
    from infrahub.database import InfrahubDatabase


def get_attribute_schema_class_for_kind(kind: str) -> type[AttributeSchema]:
    return attribute_schema_class_by_kind.get(kind, AttributeSchema)


class AttributeSchema(GeneratedAttributeSchema):
    _sort_by: list[str] = ["name"]
    _enum_class: type[enum.Enum] | None = None

    @classmethod
    def model_json_schema(cls, *args: Any, **kwargs: Any) -> dict[str, Any]:
        schema = super().model_json_schema(*args, **kwargs)

        # Build conditional schema based on attribute_schema_class_by_kind mapping
        # This override allows people using the Yaml language server to get the correct mappings
        # for the parameters when selecting the appropriate kind
        schema["allOf"] = []
        for kind, schema_class in attribute_schema_class_by_kind.items():
            schema["allOf"].append(
                {
                    "if": {"properties": {"kind": {"const": kind}}},
                    "then": {"properties": {"parameters": {"$ref": f"#/definitions/{schema_class.__name__}"}}},
                }
            )

        return schema

    @property
    def is_attribute(self) -> bool:
        return True

    @property
    def is_relationship(self) -> bool:
        return False

    @property
    def is_deprecated(self) -> bool:
        return bool(self.deprecation)

    def get_id(self) -> str:
        if self.id is None:
            raise InitializationError("The attribute schema has not been saved yet and doesn't have an id")
        return self.id

    def to_dict(self) -> dict:
        data = self.model_dump(exclude_unset=True, exclude_none=True)
        for field_name, value in data.items():
            if isinstance(value, Enum):
                data[field_name] = value.value
        return data

    @field_validator("kind")
    @classmethod
    def kind_options(cls, v: str) -> str:
        if v not in ATTRIBUTE_KIND_LABELS:
            raise ValueError(f"Only valid Attribute Kind are : {ATTRIBUTE_KIND_LABELS} ")
        return v

    @model_validator(mode="before")
    @classmethod
    def validate_dropdown_choices(cls, values: Any) -> Any:
        """Validate that choices are defined for a dropdown but not for other kinds."""
        if isinstance(values, dict):
            kind = values.get("kind")
            choices = values.get("choices")
        elif isinstance(values, AttributeSchema):
            kind = values.kind
            choices = values.choices
        else:
            return values
        if kind != "Dropdown" and choices:
            raise ValueError(f"Can only specify 'choices' for kind=Dropdown: {kind}")

        if kind == "Dropdown" and not choices:
            raise ValueError("The property 'choices' is required for kind=Dropdown")

        return values

    @field_validator("parameters", mode="before")
    @classmethod
    def set_parameters_type(cls, value: Any, info: ValidationInfo) -> Any:
        """Override parameters class if using base AttributeParameters class and should be using a subclass"""
        kind = info.data["kind"]
        expected_parameters_class = get_attribute_parameters_class_for_kind(kind=kind)
        if value is None:
            return expected_parameters_class()
        if not isinstance(value, expected_parameters_class) and isinstance(value, AttributeParameters):
            return expected_parameters_class(**value.model_dump())
        return value

    @model_validator(mode="after")
    def validate_parameters(self) -> Self:
        if isinstance(self.parameters, NumberPoolParameters) and not self.kind == "NumberPool":
            raise ValueError(f"NumberPoolParameters can't be used as parameters for {self.kind}")

        if isinstance(self.parameters, TextAttributeParameters) and self.kind not in ["Text", "TextArea"]:
            raise ValueError(f"TextAttributeParameters can't be used as parameters for {self.kind}")

        return self

    def get_class(self) -> type[BaseAttribute]:
        return ATTRIBUTE_TYPES[self.kind].get_infrahub_class()

    @property
    def uses_enum_class(self) -> bool:
        return bool(self.enum) and config.SETTINGS.experimental_features.graphql_enums

    def get_branch(self) -> BranchSupportType:
        if not self.branch:
            raise ValueError("branch hasn't been defined yet")
        return self.branch

    def get_enum_class(self) -> type[enum.Enum]:
        if not self.enum:
            raise ValueError(f"{self.name} is not an Enum")
        if not self._enum_class:
            self._enum_class = generate_python_enum(name=f"{self.name.title()}Enum", options=self.enum)
        return self._enum_class

    def convert_value_to_enum(self, value: Any) -> enum.Enum | None:
        if isinstance(value, enum.Enum) or value is None:
            return value
        enum_class = self.get_enum_class()
        return enum_class(value)

    def convert_enum_to_value(self, data: Any) -> Any:
        if isinstance(data, list):
            value = [self.convert_enum_to_value(element) for element in data]
            return [element.value if isinstance(element, enum.Enum) else element for element in value]
        if isinstance(data, enum.Enum):
            return data.value
        return data

    def update_from_generic(self, other: AttributeSchema) -> None:
        fields_to_exclude = ("id", "order_weight", "branch", "inherited")
        for name in self.model_fields:
            if name in fields_to_exclude:
                continue
            if getattr(self, name) != getattr(other, name):
                setattr(self, name, getattr(other, name))

    def to_node(self) -> dict[str, Any]:
        fields_to_exclude = {"id", "state", "filters"}
        fields_to_json = {"computed_attribute", "parameters"}
        data = self.model_dump(exclude=fields_to_exclude | fields_to_json)

        for field_name in fields_to_json:
            if field := getattr(self, field_name):
                data[field_name] = {"value": field.model_dump()}
            else:
                data[field_name] = None

        return data

    def get_regex(self) -> str | None:
        return self.regex

    def get_min_length(self) -> int | None:
        return self.min_length

    def get_max_length(self) -> int | None:
        return self.max_length

    async def get_query_filter(
        self,
        name: str,
        filter_name: str,
        branch: Branch | None = None,
        filter_value: str | int | bool | list | None = None,
        include_match: bool = True,
        param_prefix: str | None = None,
        db: InfrahubDatabase | None = None,
        partial_match: bool = False,
        support_profiles: bool = False,
    ) -> tuple[list[QueryElement], dict[str, Any], list[str]]:
        if self.enum:
            filter_value = self.convert_enum_to_value(filter_value)

        return await default_attribute_query_filter(
            name=name,
            attribute_kind=self.kind,
            filter_name=filter_name,
            branch=branch,
            filter_value=filter_value,
            include_match=include_match,
            param_prefix=param_prefix,
            db=db,
            partial_match=partial_match,
            support_profiles=support_profiles,
        )


class NumberPoolSchema(AttributeSchema):
    parameters: NumberPoolParameters = Field(
        default_factory=NumberPoolParameters,
        description="Extra parameters specific to NumberPool attributes",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )


class TextAttributeSchema(AttributeSchema):
    parameters: TextAttributeParameters = Field(
        default_factory=TextAttributeParameters,
        description="Extra parameters specific to text attributes",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )

    @model_validator(mode="after")
    def reconcile_parameters(self) -> Self:
        if self.regex != self.parameters.regex:
            final_regex = self.parameters.regex if self.parameters.regex is not None else self.regex
            self.regex = self.parameters.regex = final_regex
        if self.min_length != self.parameters.min_length:
            final_min_length = self.parameters.min_length if self.parameters.min_length is not None else self.min_length
            self.min_length = self.parameters.min_length = final_min_length
        if self.max_length != self.parameters.max_length:
            final_max_length = self.parameters.max_length if self.parameters.max_length is not None else self.max_length
            self.max_length = self.parameters.max_length = final_max_length
        return self

    def get_regex(self) -> str | None:
        return self.parameters.regex

    def get_min_length(self) -> int | None:
        return self.parameters.min_length

    def get_max_length(self) -> int | None:
        return self.parameters.max_length


class NumberAttributeSchema(AttributeSchema):
    parameters: NumberAttributeParameters = Field(
        default_factory=NumberAttributeParameters,
        description="Extra parameters specific to number attributes",
        json_schema_extra={"update": UpdateSupport.VALIDATE_CONSTRAINT.value},
    )


attribute_schema_class_by_kind: dict[str, type[AttributeSchema]] = {
    "NumberPool": NumberPoolSchema,
    "Text": TextAttributeSchema,
    "TextArea": TextAttributeSchema,
    "Number": NumberAttributeSchema,
}



from infrahub.core.node import Node

from .constants import PrefixMemberType


def get_prefix_space(ip_prefix: Node) -> int:
    prefix_space = ip_prefix.prefix.num_addresses  # type: ignore[attr-defined]

    # Non-RFC3021 subnet
    if (
        ip_prefix.member_type.value == PrefixMemberType.ADDRESS.value  # type: ignore[attr-defined]
        and ip_prefix.prefix.version == 4  # type: ignore[attr-defined]
        and ip_prefix.prefix.prefixlen < 31  # type: ignore[attr-defined]
        and not ip_prefix.is_pool.value  # type: ignore[attr-defined]
    ):
        prefix_space -= 2

    return prefix_space



from __future__ import annotations

import asyncio
import ssl
from typing import TYPE_CHECKING, Awaitable, Callable, MutableMapping, TypeVar

import nats
import ujson
from infrahub_sdk.uuidt import UUIDT
from opentelemetry import context, propagate, trace
from opentelemetry.instrumentation.utils import is_instrumentation_enabled

from infrahub import config
from infrahub.components import ComponentType
from infrahub.log import clear_log_context, get_log_data
from infrahub.message_bus import InfrahubMessage, Meta, messages
from infrahub.message_bus.operations import execute_message
from infrahub.services.adapters.message_bus import InfrahubMessageBus
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from infrahub.config import BrokerSettings
    from infrahub.message_bus.types import MessageTTL
    from infrahub.services import InfrahubServices

MessageFunction = Callable[[InfrahubMessage], Awaitable[None]]
ResponseClass = TypeVar("ResponseClass")

publish_tasks = set()


async def _add_request_id(message: InfrahubMessage) -> None:
    log_data = get_log_data()
    message.meta.request_id = log_data.get("request_id", "")


class NATSMessageBus(InfrahubMessageBus):
    def __init__(self, component_type: ComponentType, settings: BrokerSettings | None = None) -> None:
        self.settings = settings or config.SETTINGS.broker

        self.service: InfrahubServices
        self.connection: nats.NATS
        self.jetstream: nats.js.JetStreamContext
        self.callback_queue: nats.js.api.StreamInfo
        self.events_queue: nats.js.api.StreamInfo
        self.message_enrichers: list[MessageFunction] = []

        self.loop = asyncio.get_running_loop()
        self.futures: MutableMapping[str, asyncio.Future] = {}

        self.component_type: ComponentType = component_type

    @classmethod
    async def new(cls, component_type: ComponentType, settings: BrokerSettings | None = None) -> NATSMessageBus:
        message_bus = cls(component_type=component_type, settings=settings)
        await message_bus._initialize()
        return message_bus

    async def _initialize(self) -> None:
        tls_context = None
        if self.settings.tls_enabled:
            tls_context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
            if self.settings.tls_ca_file:
                tls_context.load_verify_locations(cafile=self.settings.tls_ca_file)
            if self.settings.tls_insecure:
                tls_context.check_hostname = False
                tls_context.verify_mode = ssl.CERT_NONE

        self.connection = await nats.connect(
            f"nats://{self.settings.address}:{self.settings.service_port}",
            user=self.settings.username,
            password=self.settings.password,
            tls=tls_context,
        )

        self.jetstream = self.connection.jetstream()

        if self.component_type == ComponentType.API_SERVER:
            await self._initialize_api_server()
        elif self.component_type == ComponentType.GIT_AGENT:
            await self._initialize_git_worker()

    async def shutdown(self) -> None:
        await self.connection.drain()

    async def on_callback(self, message: nats.aio.msg.Msg) -> None:
        if is_instrumentation_enabled() and message.headers:
            ctx = propagate.extract(message.headers)
            token = context.attach(ctx)

        try:
            with trace.get_tracer(__name__).start_as_current_span("on_callback") as span:
                span.set_attribute("routing_key", message.subject)

                if message.headers and "correlation_id" in message.headers:
                    span.set_attribute("correlation_id", message.headers["correlation_id"])
                    future: asyncio.Future = self.futures.pop(message.headers["correlation_id"])

                    if future:
                        future.set_result(message)
                        return

                clear_log_context()
                if message.subject in messages.MESSAGE_MAP:
                    await execute_message(routing_key=message.subject, message_body=message.data, service=self.service)
                else:
                    self.service.log.error("Invalid message received", message=f"{message!r}")
        finally:
            if is_instrumentation_enabled() and message.headers:
                context.detach(token)

    async def on_message(self, message: nats.aio.msg.Msg) -> None:
        if is_instrumentation_enabled() and message.headers:
            ctx = propagate.extract(message.headers)
            token = context.attach(ctx)

        try:
            with trace.get_tracer(__name__).start_as_current_span("on_message") as span:
                span.set_attribute("routing_key", message.subject)

                clear_log_context()
                if message.subject in messages.MESSAGE_MAP:
                    delay = await execute_message(
                        routing_key=message.subject, message_body=message.data, service=self.service
                    )
                    if delay:
                        return await message.nak(delay / 1000)
                else:
                    self.service.log.error("Invalid message received", message=f"{message!r}")

                return await message.ack()
        finally:
            if is_instrumentation_enabled() and message.headers:
                context.detach(token)

    async def _subscribe_events(self, events: list[str], identity: str) -> None:
        for subject in events:
            await self.jetstream.subscribe(
                subject=subject,
                stream=f"{self.settings.namespace}-events",
                config=nats.js.api.ConsumerConfig(ack_policy=nats.js.api.AckPolicy.NONE, description=identity),
                cb=self.on_callback,
            )

    async def _setup_callback(self, identity: str) -> None:
        self.callback_queue = await self.jetstream.add_stream(
            name=f"{self.settings.namespace}-callback-{WORKER_IDENTITY}",
            retention=nats.js.api.RetentionPolicy.LIMITS,
        )

        await self.jetstream.subscribe(
            subject="*",
            stream=f"{self.settings.namespace}-callback-{WORKER_IDENTITY}",
            cb=self.on_callback,
            config=nats.js.api.ConsumerConfig(ack_policy=nats.js.api.AckPolicy.NONE, description=identity),
        )

    async def _initialize_api_server(self) -> None:
        self.events_queue = await self.jetstream.add_stream(
            name=f"{self.settings.namespace}-events",
            subjects=self.event_bindings,
            retention=nats.js.api.RetentionPolicy.INTEREST,
        )

        await self.jetstream.add_stream(
            name=f"{self.settings.namespace}-rpcs",
            subjects=self.worker_bindings,
            retention=nats.js.api.RetentionPolicy.WORK_QUEUE,
        )

        await self._subscribe_events(self.event_bindings, f"api-worker-{WORKER_IDENTITY}")

        await self._setup_callback(f"api-worker-{WORKER_IDENTITY}")

        self.message_enrichers.append(_add_request_id)

    async def _initialize_git_worker(self) -> None:
        bindings = self.event_bindings + self.broadcasted_event_bindings
        await self._subscribe_events(bindings, f"git-worker-{WORKER_IDENTITY}")

        consumer_config = nats.js.api.ConsumerConfig(
            ack_policy=nats.js.api.AckPolicy.EXPLICIT,
            max_deliver=self.settings.maximum_message_retries,
            ack_wait=self.DELIVER_TIMEOUT,
            # Does not work as expected, must switch to pull-based consumer...
            # max_ack_pending=self.settings.maximum_concurrent_messages,
            # flow_control=True,
            # idle_heartbeat=5.0,  # default value
            filter_subjects=bindings,
            durable_name="git-workers",
            deliver_group="git-workers",
            deliver_subject=self.connection.new_inbox(),
        )

        try:
            await self.jetstream.add_consumer(stream=f"{self.settings.namespace}-rpcs", config=consumer_config)
        except nats.js.errors.BadRequestError as exc:
            if exc.err_code != 10013:  # consumer name already in use
                raise

        for subject in bindings:
            await self.jetstream.subscribe(
                subject=subject,
                queue="git-workers",
                stream=f"{self.settings.namespace}-rpcs",
                config=consumer_config,
                cb=self.on_message,
                manual_ack=True,
            )

        await self._setup_callback(f"git-worker-{WORKER_IDENTITY}")

    async def _publish_with_delay(self, message: InfrahubMessage, routing_key: str, delay: MessageTTL) -> None:
        await asyncio.sleep(delay.value / 1000)
        await self.publish(message, routing_key)

    async def publish(
        self, message: InfrahubMessage, routing_key: str, delay: MessageTTL | None = None, is_retry: bool = False
    ) -> None:
        with trace.get_tracer(__name__).start_as_current_span("publish_message") as span:
            span.set_attribute("routing_key", routing_key)

            if delay:
                if is_retry:
                    # Delayed retries are directly handled in the callback using Nack
                    return
                # Use asyncio task for delayed publish since NATS does not support that out of the box
                task = asyncio.create_task(self._publish_with_delay(message, routing_key, delay))
                publish_tasks.add(task)
                task.add_done_callback(publish_tasks.discard)
                return

            for enricher in self.message_enrichers:
                await enricher(message)
            message.assign_priority(priority=messages.message_priority(routing_key=routing_key))

            headers = {}
            if message.meta.correlation_id:
                headers["correlation_id"] = message.meta.correlation_id
            if message.meta.reply_to:
                headers["reply_to"] = message.meta.reply_to
            if message.meta.expiration:
                headers["expiration"] = str(message.meta.expiration)

            if message.meta.headers:
                # Filter None and non-string values
                for k, v in message.meta.headers.items():
                    if v:
                        headers[k] = str(v)

            if is_instrumentation_enabled():
                propagate.inject(headers)

            await self.jetstream.publish(
                subject=routing_key,
                payload=message.body,
                headers=headers,  # None value throws exception
            )

    async def reply(self, message: InfrahubMessage, routing_key: str) -> None:
        headers = {}
        if message.meta.correlation_id:
            headers["correlation_id"] = message.meta.correlation_id
        if message.meta.reply_to:
            headers["reply_to"] = message.meta.reply_to
        if message.meta.expiration:
            headers["expiration"] = str(message.meta.expiration)
        if message.meta.headers:
            # Filter None and non-string values
            for k, v in message.meta.headers.items():
                if v:
                    headers[k] = str(v)

        if is_instrumentation_enabled():
            propagate.inject(headers)

        await self.jetstream.publish(
            subject=routing_key,
            payload=message.body,
            headers=headers,
        )

    async def rpc(self, message: InfrahubMessage, response_class: type[ResponseClass]) -> ResponseClass:
        correlation_id = str(UUIDT())

        future = self.loop.create_future()
        self.futures[correlation_id] = future

        log_data = get_log_data()
        request_id = log_data.get("request_id", "")
        message.meta = Meta(
            request_id=request_id, correlation_id=correlation_id, reply_to=self.callback_queue.config.name
        )

        await self.service.message_bus.send(message=message)

        response: nats.aio.msg.Msg = await future
        data = ujson.loads(response.data)
        return response_class(**data)



from infrahub.core.node.constraints.attribute_uniqueness import NodeAttributeUniquenessConstraint
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class NodeAttributeUniquenessConstraintDependency(DependencyBuilder[NodeAttributeUniquenessConstraint]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> NodeAttributeUniquenessConstraint:
        return NodeAttributeUniquenessConstraint(db=context.db, branch=context.branch)



from infrahub.auth import AccountSession
from infrahub.core.branch import Branch
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import PermissionDeniedError
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.initialization import GraphqlParams

from .interface import CheckerResolution, GraphQLQueryPermissionCheckerInterface


class GraphQLQueryPermissionChecker:
    def __init__(self, sub_checkers: list[GraphQLQueryPermissionCheckerInterface]) -> None:
        self.sub_checkers = sub_checkers

    async def check(
        self,
        db: InfrahubDatabase,
        account_session: AccountSession,
        analyzed_query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,
        branch: Branch,
    ) -> None:
        for sub_checker in self.sub_checkers:
            if await sub_checker.supports(db=db, account_session=account_session, branch=branch):
                resolution = await sub_checker.check(
                    db=db,
                    account_session=account_session,
                    analyzed_query=analyzed_query,
                    query_parameters=query_parameters,
                    branch=branch,
                )
                if resolution == CheckerResolution.TERMINATE:
                    return
        raise PermissionDeniedError("The current account is not authorized to perform this operation")



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from neo4j.graph import Path as Neo4jPath


class InvalidCypherPathError(Exception):
    def __init__(self, cypher_path: Neo4jPath) -> None:
        self.cypher_path = cypher_path

    def __str__(self) -> str:
        path_repr = ""
        for relationship in self.cypher_path.relationships:
            path_repr += f"(type={relationship.type} {relationship.start_node}-{relationship.end_node})"
        return f"Cannot parse cypher path {path_repr}"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Argument, Boolean, DateTime, Field, Int, List, NonNull, ObjectType, String
from infrahub_sdk.utils import extract_fields_first_node

from infrahub.exceptions import ValidationError
from infrahub.graphql.types.event import EventNodes, EventTypeFilter
from infrahub.task_manager.event import PrefectEvent
from infrahub.task_manager.models import InfrahubEventFilter

if TYPE_CHECKING:
    from datetime import datetime

    from graphql import GraphQLResolveInfo


class Events(ObjectType):
    edges = List(NonNull(EventNodes), required=True)
    count = Int(required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        limit: int = 10,
        has_children: bool | None = None,
        level: int | None = None,
        offset: int | None = None,
        account__ids: list[str] | None = None,
        ids: list[str] | None = None,
        branches: list[str] | None = None,
        event_type: list[str] | None = None,
        event_type_filter: dict[str, Any] | None = None,
        related_node__ids: list[str] | None = None,
        primary_node__ids: list[str] | None = None,
        parent__ids: list[str] | None = None,
        since: datetime | None = None,
        until: datetime | None = None,
    ) -> dict[str, Any]:
        ids = ids or []
        if limit > 50:
            # Prefect restricts this to 50
            raise ValidationError(input_value="The parameter 'limit' can't be above 50")

        event_filter = InfrahubEventFilter.from_filters(
            ids=ids,
            branches=branches,
            account__ids=account__ids,
            has_children=has_children,
            event_type=event_type,
            event_type_filter=event_type_filter,
            related_node__ids=related_node__ids,
            primary_node__ids=primary_node__ids,
            parent__ids=parent__ids,
            since=since,
            until=until,
            level=level,
        )

        return await Events.query(
            info=info,
            event_filter=event_filter,
            limit=limit,
            offset=offset,
        )

    @classmethod
    async def query(
        cls,
        info: GraphQLResolveInfo,
        event_filter: InfrahubEventFilter,
        limit: int,
        offset: int | None = None,
    ) -> dict[str, Any]:
        fields = await extract_fields_first_node(info)

        prefect_tasks = await PrefectEvent.query(
            fields=fields,
            event_filter=event_filter,
            limit=limit,
            offset=offset,
        )
        return {
            "count": prefect_tasks.get("count", 0),
            "edges": prefect_tasks.get("edges", []),
        }


Event = Field(
    Events,
    limit=Int(required=False),
    offset=Int(required=False),
    level=Int(required=False),
    has_children=Boolean(required=False, description="Filter events based on if they can have children or not"),
    event_type=List(NonNull(String), description="Filter events that match a specific type"),
    event_type_filter=Argument(EventTypeFilter, required=False, description="Filters specific to a given event_type"),
    primary_node__ids=List(
        NonNull(String), description="Filter events where the primary node id is within indicated node ids"
    ),
    related_node__ids=List(
        NonNull(String), description="Filter events where the related node ids are within indicated node ids"
    ),
    parent__ids=List(
        NonNull(String), description="Search events that has any of the indicated event ids listed as parents"
    ),
    since=DateTime(required=False, description="Search events since this timestamp, defaults to 180 days back"),
    until=DateTime(required=False, description="Search events until this timestamp, defaults the current time"),
    branches=List(NonNull(String), required=False, description="Filter the query to specific branches"),
    account__ids=List(NonNull(String), required=False, description="Filter the query to specific accounts"),
    ids=List(NonNull(String)),
    resolver=Events.resolve,
    required=True,
)



from typing import AsyncGenerator

from infrahub.core.constants import NULL_VALUE, DiffAction, RelationshipCardinality
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.schema import MainSchemaTypes
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import SchemaNotFoundError
from infrahub.types import ATTRIBUTE_PYTHON_TYPES

from ..model.path import (
    ConflictSelection,
    EnrichedDiffAttribute,
    EnrichedDiffConflict,
    EnrichedDiffProperty,
    EnrichedDiffRoot,
    EnrichedDiffSingleRelationship,
)
from .model import (
    AttributeMergeDict,
    AttributePropertyMergeDict,
    NodeMergeDict,
    PropertyMergeDict,
    RelationshipMergeDict,
    RelationshipPropertyMergeDict,
)

Primitives = str | bool | int | float


class DiffMergeSerializer:
    def __init__(self, db: InfrahubDatabase, max_batch_size: int) -> None:
        self.db = db
        self.max_batch_size = max_batch_size
        self._attribute_type_cache: dict[tuple[str, str], type] = {}
        self._source_branch_name: str | None = None
        self._target_branch_name: str | None = None
        # {(node_id, relationship_id, peer_id)}
        self._conflicted_cardinality_one_relationships: set[tuple[str, str, str]] = set()

    def _reset_caches(self) -> None:
        self._attribute_type_cache = {}
        self._conflicted_cardinality_one_relationships = set()

    @property
    def source_branch_name(self) -> str:
        if self._source_branch_name is None:
            raise RuntimeError("source_branch_name not set")
        return self._source_branch_name

    @property
    def target_branch_name(self) -> str:
        if self._target_branch_name is None:
            raise RuntimeError("target_branch_name not set")
        return self._target_branch_name

    def _get_schema(self, kind: str, branch_name: str) -> MainSchemaTypes:
        schema_branch = self.db.schema.get_schema_branch(name=branch_name)
        return schema_branch.get(name=kind, duplicate=False)

    def _get_action(self, action: DiffAction, conflict: EnrichedDiffConflict | None) -> DiffAction:
        if not conflict:
            return action
        if conflict.selected_branch is ConflictSelection.BASE_BRANCH:
            return conflict.base_branch_action
        if conflict.selected_branch is ConflictSelection.DIFF_BRANCH:
            return conflict.diff_branch_action
        raise ValueError(f"conflict {conflict.uuid} does not have a branch selection")

    def _to_action_str(self, action: DiffAction) -> str:
        return str(action.value).upper()

    def _get_property_type_for_attribute_value(self, schema_kind: str, attribute_name: str) -> type:
        cache_key = (schema_kind, attribute_name)
        if cache_key in self._attribute_type_cache:
            return self._attribute_type_cache[cache_key]
        try:
            node_schema = self._get_schema(kind=schema_kind, branch_name=self.source_branch_name)
            attribute_schema = node_schema.get_attribute(name=attribute_name)
        except (SchemaNotFoundError, ValueError):
            node_schema = self._get_schema(kind=schema_kind, branch_name=self.target_branch_name)
            attribute_schema = node_schema.get_attribute(name=attribute_name)
        python_type = ATTRIBUTE_PYTHON_TYPES[attribute_schema.kind]
        final_python_type: type = str
        if python_type in (str, int, float, bool):
            final_python_type = python_type
        self._attribute_type_cache[cache_key] = final_python_type
        return final_python_type

    def _convert_property_value(
        self, property_type: DatabaseEdgeType, raw_value: str | None, value_type: type | None = None
    ) -> Primitives | None:
        # peer IDs are strings
        if property_type in (DatabaseEdgeType.HAS_OWNER, DatabaseEdgeType.HAS_SOURCE, DatabaseEdgeType.IS_RELATED):
            return raw_value
        # these are boolean
        if (property_type in (DatabaseEdgeType.IS_VISIBLE, DatabaseEdgeType.IS_PROTECTED)) and isinstance(
            raw_value, str
        ):
            return raw_value.lower() == "true"
        # this must be HAS_VALUE
        if raw_value in (None, NULL_VALUE):
            return NULL_VALUE
        if value_type:
            if value_type is bool and isinstance(raw_value, str):
                return raw_value.lower() == "true"
            return value_type(raw_value)
        return raw_value

    def _cache_conflicted_cardinality_one_relationships(self, diff: EnrichedDiffRoot) -> None:
        for node in diff.nodes:
            for rel in node.relationships:
                if rel.cardinality is not RelationshipCardinality.ONE:
                    continue
                for element in rel.relationships:
                    if element.conflict is None:
                        continue
                    for prop in element.properties:
                        if prop.property_type is not DatabaseEdgeType.IS_RELATED:
                            continue
                        if prop.previous_value:
                            self._conflicted_cardinality_one_relationships.add(
                                (node.uuid, rel.identifier, prop.previous_value)
                            )
                        if prop.new_value:
                            self._conflicted_cardinality_one_relationships.add(
                                (node.uuid, rel.identifier, prop.new_value)
                            )

    async def serialize_diff(
        self, diff: EnrichedDiffRoot
    ) -> AsyncGenerator[
        tuple[list[NodeMergeDict], list[AttributePropertyMergeDict | RelationshipPropertyMergeDict]], None
    ]:
        self._reset_caches()
        self._source_branch_name = diff.diff_branch_name
        self._target_branch_name = diff.base_branch_name
        self._cache_conflicted_cardinality_one_relationships(diff=diff)
        serialized_node_diffs = []
        serialized_property_diffs: list[AttributePropertyMergeDict | RelationshipPropertyMergeDict] = []
        for node in diff.nodes:
            # if there is a node-level conflict and the base branch version is selected
            # then we don't need to alter this node during the merge
            if node.conflict and node.conflict.selected_branch is ConflictSelection.BASE_BRANCH:
                continue
            node_action = self._get_action(action=node.action, conflict=node.conflict)
            if node_action is DiffAction.REMOVED:
                serialized_node_diffs.append(
                    NodeMergeDict(
                        uuid=node.uuid,
                        action=self._to_action_str(action=node_action),
                        attributes=[],
                        relationships=[],
                    )
                )
                continue
            serial_attr_diffs = []
            for attr_diff in node.attributes:
                serial_attr_diff, attribute_property_diff = self._serialize_attribute(
                    attribute_diff=attr_diff, node_uuid=node.uuid, node_kind=node.kind
                )
                if serial_attr_diff:
                    serial_attr_diffs.append(serial_attr_diff)
                serialized_property_diffs.append(attribute_property_diff)
            relationship_diffs: list[RelationshipMergeDict] = []
            for rel_diff in node.relationships:
                for relationship_element_diff in rel_diff.relationships:
                    element_diffs, relationship_property_diffs = self._serialize_relationship_element(
                        relationship_diff=relationship_element_diff,
                        relationship_identifier=rel_diff.identifier,
                        node_uuid=node.uuid,
                    )
                    relationship_diffs.extend(element_diffs)
                    serialized_property_diffs.extend(relationship_property_diffs)
            if node_action in (DiffAction.ADDED, DiffAction.REMOVED) or serial_attr_diffs or relationship_diffs:
                serialized_node_diffs.append(
                    NodeMergeDict(
                        uuid=node.uuid,
                        action=self._to_action_str(action=node_action),
                        attributes=serial_attr_diffs,
                        relationships=relationship_diffs,
                    )
                )
            if len(serialized_node_diffs) == self.max_batch_size:
                yield (serialized_node_diffs, serialized_property_diffs)
                serialized_node_diffs, serialized_property_diffs = [], []
        yield (serialized_node_diffs, serialized_property_diffs)

    def _get_property_actions_and_values(
        self, property_diff: EnrichedDiffProperty, python_value_type: type
    ) -> list[tuple[DiffAction, Primitives]]:
        action = property_diff.action
        new_value = property_diff.new_value
        if property_diff.conflict and property_diff.conflict.selected_branch is ConflictSelection.BASE_BRANCH:
            action = property_diff.conflict.base_branch_action
            if property_diff.conflict.base_branch_value:
                new_value = property_diff.conflict.base_branch_value
        actions = [action]
        if property_diff.action is DiffAction.UPDATED:
            actions = [DiffAction.ADDED, DiffAction.REMOVED]
        actions_and_values: list[tuple[DiffAction, Primitives]] = []
        for action in actions:
            if action not in (DiffAction.ADDED, DiffAction.REMOVED):
                continue
            if action is DiffAction.REMOVED and new_value == NULL_VALUE:
                actions_and_values.append((DiffAction.ADDED, NULL_VALUE))
                continue
            if action is DiffAction.ADDED:
                raw_value = new_value
            else:
                raw_value = property_diff.previous_value
            final_value = self._convert_property_value(
                property_type=property_diff.property_type, raw_value=raw_value, value_type=python_value_type
            )
            if final_value is not None:
                actions_and_values.append((action, final_value))
        return actions_and_values

    def _serialize_attribute(
        self, attribute_diff: EnrichedDiffAttribute, node_uuid: str, node_kind: str
    ) -> tuple[AttributeMergeDict | None, AttributePropertyMergeDict]:
        prop_dicts: list[PropertyMergeDict] = []
        python_type = self._get_property_type_for_attribute_value(
            schema_kind=node_kind, attribute_name=attribute_diff.name
        )
        for property_diff in attribute_diff.properties:
            actions_and_values = self._get_property_actions_and_values(
                property_diff=property_diff, python_value_type=python_type
            )
            for action, value in actions_and_values:
                # we only delete attributes when the whole attribute is deleted
                if action is DiffAction.REMOVED and attribute_diff.action is not DiffAction.REMOVED:
                    continue
                prop_dicts.append(
                    PropertyMergeDict(
                        property_type=property_diff.property_type.value,
                        action=self._to_action_str(action=action),
                        value=value,
                    )
                )
        attr_dict = None
        if attribute_diff.action in (DiffAction.ADDED, DiffAction.REMOVED):
            attr_dict = AttributeMergeDict(
                name=attribute_diff.name,
                action=self._to_action_str(action=attribute_diff.action),
            )
        attr_prop_dict = AttributePropertyMergeDict(
            node_uuid=node_uuid, attribute_name=attribute_diff.name, properties=prop_dicts
        )
        return attr_dict, attr_prop_dict

    def _get_default_property_merge_dicts(self, action: DiffAction) -> dict[DatabaseEdgeType, PropertyMergeDict]:
        # start with default values for IS_VISIBLE and IS_PROTECTED b/c we always want to update them during a merge
        return {
            DatabaseEdgeType.IS_VISIBLE: PropertyMergeDict(
                property_type=DatabaseEdgeType.IS_VISIBLE.value,
                action=self._to_action_str(action),
                value=None,
            ),
            DatabaseEdgeType.IS_PROTECTED: PropertyMergeDict(
                property_type=DatabaseEdgeType.IS_PROTECTED.value,
                action=self._to_action_str(action),
                value=None,
            ),
        }

    def _get_actions_and_peers(self, relationship_diff: EnrichedDiffSingleRelationship) -> list[tuple[DiffAction, str]]:
        is_related_prop = [p for p in relationship_diff.properties if p.property_type is DatabaseEdgeType.IS_RELATED][0]
        actions_and_values = self._get_property_actions_and_values(property_diff=is_related_prop, python_value_type=str)
        actions_and_peers: list[tuple[DiffAction, str]] = []
        for action, peer_id in actions_and_values:
            if action is DiffAction.ADDED:
                actions_and_peers.append((DiffAction.ADDED, str(peer_id)))
            elif action is DiffAction.REMOVED:
                actions_and_peers.append((DiffAction.REMOVED, str(peer_id)))

        conflict = relationship_diff.conflict
        if (
            conflict
            and conflict.selected_branch
            and conflict.selected_branch is ConflictSelection.DIFF_BRANCH
            and conflict.base_branch_value
            and conflict.base_branch_action in (DiffAction.ADDED, DiffAction.UPDATED)
        ):
            actions_and_peers.append((DiffAction.REMOVED, conflict.base_branch_value))
        return actions_and_peers

    def _serialize_relationship_element(
        self, relationship_diff: EnrichedDiffSingleRelationship, relationship_identifier: str, node_uuid: str
    ) -> tuple[list[RelationshipMergeDict], list[RelationshipPropertyMergeDict]]:
        # if there is a relationship-element conflict and we are keeping the base branch version
        # then we do not need to do anything special
        if relationship_diff.conflict and relationship_diff.conflict.selected_branch is ConflictSelection.BASE_BRANCH:
            return ([], [])
        relationship_dicts = []
        actions_and_peers = self._get_actions_and_peers(relationship_diff=relationship_diff)
        added_peer_ids = [peer_id for action, peer_id in actions_and_peers if action is DiffAction.ADDED]
        removed_peer_ids = [peer_id for action, peer_id in actions_and_peers if action is DiffAction.REMOVED]
        for action, peer_id in actions_and_peers:
            if (
                peer_id
                and (peer_id, relationship_identifier, node_uuid) not in self._conflicted_cardinality_one_relationships
            ):
                relationship_dicts.append(
                    RelationshipMergeDict(
                        peer_id=peer_id, name=relationship_identifier, action=self._to_action_str(action=action)
                    )
                )
        relationship_property_dicts = self._serialize_relationship_properties(
            node_uuid=node_uuid,
            relationship_identifier=relationship_identifier,
            relationship_diff_properties=relationship_diff.properties,
            added_peer_ids=added_peer_ids,
            removed_peer_ids=removed_peer_ids,
            unchanged_peer_id=relationship_diff.peer_id,
        )
        return relationship_dicts, relationship_property_dicts

    def _serialize_relationship_properties(
        self,
        node_uuid: str,
        relationship_identifier: str,
        relationship_diff_properties: set[EnrichedDiffProperty],
        added_peer_ids: list[str],
        removed_peer_ids: list[str],
        unchanged_peer_id: str,
    ) -> list[RelationshipPropertyMergeDict]:
        added_property_dicts = {}
        removed_property_dicts = {}
        if added_peer_ids:
            added_property_dicts = self._get_default_property_merge_dicts(action=DiffAction.ADDED)
        if removed_peer_ids:
            removed_property_dicts = self._get_default_property_merge_dicts(action=DiffAction.REMOVED)
        for property_diff in relationship_diff_properties:
            if property_diff.property_type is DatabaseEdgeType.IS_RELATED:
                # handled above
                continue
            python_value_type: type = str
            if property_diff.property_type in (DatabaseEdgeType.IS_VISIBLE, DatabaseEdgeType.IS_PROTECTED):
                python_value_type = bool
            actions_and_values = self._get_property_actions_and_values(
                property_diff=property_diff, python_value_type=python_value_type
            )
            for action, value in actions_and_values:
                property_dict = PropertyMergeDict(
                    property_type=property_diff.property_type.value,
                    action=self._to_action_str(action=action),
                    value=value,
                )
                if action is DiffAction.ADDED:
                    added_property_dicts[property_diff.property_type] = property_dict
                elif action is DiffAction.REMOVED:
                    removed_property_dicts[property_diff.property_type] = property_dict
        relationship_property_dicts = []
        peers_and_property_dicts: list[tuple[str, dict[DatabaseEdgeType, PropertyMergeDict]]] = []
        if added_property_dicts:
            peers_and_property_dicts += [
                (peer_id, added_property_dicts) for peer_id in (added_peer_ids or [unchanged_peer_id])
            ]
        if removed_property_dicts:
            peers_and_property_dicts += [
                (peer_id, removed_property_dicts) for peer_id in (removed_peer_ids or [unchanged_peer_id])
            ]
        for peer_id, property_dicts in peers_and_property_dicts:
            if (
                peer_id
                and property_dicts
                and (peer_id, relationship_identifier, node_uuid) not in self._conflicted_cardinality_one_relationships
            ):
                relationship_property_dicts.append(
                    RelationshipPropertyMergeDict(
                        node_uuid=node_uuid,
                        relationship_id=relationship_identifier,
                        peer_uuid=peer_id,
                        properties=list(property_dicts.values()),
                    )
                )
        return relationship_property_dicts



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

from prefect.client.orchestration import PrefectClient, get_client
from prefect.events.schemas.events import Event as PrefectEventModel
from prefect.exceptions import PrefectHTTPStatusError
from pydantic import BaseModel, Field, TypeAdapter

from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.exceptions import ServiceUnavailableError
from infrahub.log import get_logger
from infrahub.utils import get_nested_dict

log = get_logger()

if TYPE_CHECKING:
    from .models import InfrahubEventFilter


class PrefectEventData(PrefectEventModel):
    def get_branch(self) -> str | None:
        for resource in self.related:
            if resource.get("prefect.resource.role") != "infrahub.branch":
                continue
            if "infrahub.resource.label" not in resource:
                continue
            if resource.get("infrahub.resource.label") == GLOBAL_BRANCH_NAME:
                return None
            return resource.get("infrahub.resource.label")
        return None

    def get_level(self) -> int:
        for resource in self.related:
            level = resource.get("infrahub.event.level")
            if level is None:
                continue
            try:
                return int(level)
            except ValueError:
                return 0

        return 0

    def get_parent(self) -> str | None:
        for resource in self.related:
            if resource.get("prefect.resource.role") != "infrahub.child_event":
                continue
            return resource.get("infrahub.event_parent.id")
        return None

    def get_primary_node(self) -> dict[str, str] | None:
        node_id = self.resource.get("infrahub.node.id")
        node_kind = self.resource.get("infrahub.node.kind")
        if node_id and node_kind:
            return {"id": node_id, "kind": node_kind}

        return None

    def get_related_nodes(self) -> list[dict[str, str]]:
        related_nodes = []
        for resource in self.related:
            if resource.get("prefect.resource.role") != "infrahub.related.node":
                continue

            node_id = resource.get("prefect.resource.id")
            node_kind = resource.get("infrahub.node.kind")
            if node_id == self.resource.get("infrahub.node.id"):
                # Don't include the primary node as a related node.
                continue
            if node_id and node_kind:
                related_nodes.append({"id": node_id, "kind": node_kind})

        return related_nodes

    def get_account_id(self) -> str | None:
        for resource in self.related:
            if resource.get("prefect.resource.role") != "infrahub.account":
                continue
            return resource.get("infrahub.resource.id")
        return None

    def has_children(self) -> bool:
        for resource in self.related:
            if resource.get("prefect.resource.role") != "infrahub.event":
                continue
            if resource.get("infrahub.event.has_children") == "true":
                return True
            return False
        return False

    def _return_node_mutation(self) -> dict[str, Any]:
        attributes = []
        relationships = []

        for resource in self.related:
            if resource.role == "infrahub.node.attribute_update" and resource.get("infrahub.attribute.name"):
                attributes.append(
                    {
                        "name": resource.get("infrahub.attribute.name", ""),
                        "kind": resource.get("infrahub.attribute.kind", ""),
                        "value": None
                        if resource.get("infrahub.attribute.value") == "NULL"
                        else resource.get("infrahub.attribute.value"),
                        "value_previous": None
                        if resource.get("infrahub.attribute.value_previous") == "NULL"
                        else resource.get("infrahub.attribute.value_previous"),
                        "action": resource.get("infrahub.attribute.action", "unchanged"),
                    }
                )
            elif resource.role == "infrahub.node.relationship_update":
                relationships.append(
                    {
                        "name": resource.get("infrahub.relationship.name"),
                        "action": resource.get("infrahub.relationship.peer_status"),
                        "peer": {
                            "id": resource.get("infrahub.relationship.peer_id"),
                            "kind": resource.get("infrahub.relationship.peer_kind"),
                        },
                    }
                )

        return {"attributes": attributes, "relationships": relationships}

    def _get_branch_name_from_resource(self) -> str:
        return self.resource.get("infrahub.branch.name") or ""

    def _return_artifact_event(self) -> dict[str, Any]:
        checksum = ""
        checksum_previous: str | None = None
        storage_id = ""
        storage_id_previous: str | None = None
        artifact_definition_id = ""
        for resource in self.related:
            if resource.role == "infrahub.artifact":
                checksum = resource.get("infrahub.artifact.checksum") or ""
                checksum_previous = resource.get("infrahub.artifact.checksum_previous")
                storage_id = resource.get("infrahub.artifact.storage_id") or ""
                storage_id_previous = resource.get("infrahub.artifact.storage_id_previous")
                artifact_definition_id = resource.get("infrahub.artifact.artifact_definition_id") or ""

        return {
            "checksum": checksum,
            "checksum_previous": checksum_previous,
            "storage_id": storage_id,
            "storage_id_previous": storage_id_previous,
            "artifact_definition_id": artifact_definition_id,
        }

    def _return_branch_created(self) -> dict[str, Any]:
        return {"created_branch": self._get_branch_name_from_resource()}

    def _return_branch_deleted(self) -> dict[str, Any]:
        return {"deleted_branch": self._get_branch_name_from_resource()}

    def _return_branch_merged(self) -> dict[str, Any]:
        return {"source_branch": self._get_branch_name_from_resource()}

    def _return_branch_rebased(self) -> dict[str, Any]:
        return {"rebased_branch": self._get_branch_name_from_resource()}

    def _return_group_event(self) -> dict[str, Any]:
        members = []
        ancestors = []

        for resource in self.related:
            if resource.role == "infrahub.group.member" and resource.get("infrahub.node.kind"):
                members.append({"id": resource.id, "kind": resource.get("infrahub.node.kind")})
            elif resource.role == "infrahub.group.ancestor" and resource.get("infrahub.node.kind"):
                ancestors.append({"id": resource.id, "kind": resource.get("infrahub.node.kind")})

        return {"members": members, "ancestors": ancestors}

    def _return_event_specifics(self) -> dict[str, Any]:
        """Return event specific data based on the type of event being processed"""

        event_specifics = {}

        match self.event:
            case "infrahub.artifact.created" | "infrahub.artifact.updated":
                event_specifics = self._return_artifact_event()
            case "infrahub.node.created" | "infrahub.node.updated" | "infrahub.node.deleted":
                event_specifics = self._return_node_mutation()
            case "infrahub.branch.created":
                event_specifics = self._return_branch_created()
            case "infrahub.branch.deleted":
                event_specifics = self._return_branch_deleted()
            case "infrahub.branch.merged":
                event_specifics = self._return_branch_merged()
            case "infrahub.branch.rebased":
                event_specifics = self._return_branch_rebased()
            case "infrahub.group.member_added" | "infrahub.group.member_removed":
                event_specifics = self._return_group_event()

        return event_specifics

    def to_graphql(self) -> dict[str, Any]:
        response = {
            "id": str(self.id),
            "event": self.event,
            "branch": self.get_branch(),
            "account_id": self.get_account_id(),
            "occurred_at": self.occurred,
            "has_children": self.has_children(),
            "payload": self.payload,
            "level": self.get_level(),
            "primary_node": self.get_primary_node(),
            "parent_id": self.get_parent(),
            "related_nodes": self.get_related_nodes(),
        }
        response.update(self._return_event_specifics())
        return response


class PrefectEventResponse(BaseModel):
    count: int = Field(..., description="Number of matching events")
    events: list[PrefectEventData] = Field(..., description="Returned events")


class PrefectEvent:
    @classmethod
    async def query_events(
        cls,
        client: PrefectClient,
        limit: int,
        filters: InfrahubEventFilter,
        offset: int | None = None,
    ) -> PrefectEventResponse:
        body = {"limit": limit, "filter": filters.model_dump(mode="json", exclude_none=True), "offset": offset}

        # Retry due to https://github.com/PrefectHQ/prefect/issues/16299
        for _ in range(1, 5):
            prefect_error: PrefectHTTPStatusError | None = None
            try:
                response = await client._client.post("/infrahub/events/filter", json=body)
                break
            except PrefectHTTPStatusError as exc:
                prefect_error = exc
                await asyncio.sleep(0.1)

        if prefect_error:
            raise ServiceUnavailableError(
                message=f"Unable to query prefect due to invalid response from the server (status_code={prefect_error.response.status_code})"
            )
        data: dict[str, Any] = response.json()

        return PrefectEventResponse(
            count=data.get("total", 0),
            events=TypeAdapter(list[PrefectEventData]).validate_python(data.get("events")),
        )

    @classmethod
    async def query(
        cls,
        fields: dict[str, Any],
        event_filter: InfrahubEventFilter,
        limit: int | None = None,
        offset: int | None = None,
    ) -> dict[str, Any]:
        nodes: list[dict] = []
        limit = limit or 50

        node_fields = get_nested_dict(nested_dict=fields, keys=["edges", "node"])

        if not node_fields:
            # This means that it's purely a count query and as such we can override the limit to avoid
            # returning data that will only be discarded
            limit = 1

        async with get_client(sync_client=False) as client:
            response = await cls.query_events(client=client, filters=event_filter, limit=limit, offset=offset)
            nodes = [{"node": event.to_graphql()} for event in response.events]

        return {"count": response.count, "edges": nodes}



from abc import ABC, abstractmethod

from graphene import InputObjectType

from infrahub.core.branch import Branch
from infrahub.core.node import Node
from infrahub.core.schema import MainSchemaTypes


class MutationNodeGetterInterface(ABC):
    @abstractmethod
    async def get_node(
        self,
        node_schema: MainSchemaTypes,
        data: InputObjectType,
        branch: Branch,
    ) -> Node | None: ...



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from infrahub import lock
from infrahub.core.manager import NodeManager
from infrahub.core.models import (
    HashableModelDiff,
    SchemaBranchDiff,
    SchemaDiff,
)
from infrahub.core.node import Node
from infrahub.core.registry import registry
from infrahub.core.schema import (
    AttributeSchema,
    GenericSchema,
    MainSchemaTypes,
    NodeSchema,
    ProfileSchema,
    RelationshipSchema,
    SchemaRoot,
    TemplateSchema,
)
from infrahub.core.utils import parse_node_kind
from infrahub.exceptions import SchemaNotFoundError
from infrahub.log import get_logger

from .constants import IGNORE_FOR_NODE
from .schema_branch import SchemaBranch

log = get_logger()

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class SchemaManager(NodeManager):
    def __init__(self) -> None:
        self._cache: dict[int, Any] = {}
        self._branches: dict[str, SchemaBranch] = {}

    def _get_from_cache(self, key: int) -> Any:
        return self._cache[key]

    def set(self, name: str, schema: NodeSchema | GenericSchema, branch: str | None = None) -> int:
        branch = branch or registry.default_branch

        if branch not in self._branches:
            self._branches[branch] = SchemaBranch(cache=self._cache, name=branch)

        self._branches[branch].set(name=name, schema=schema)

        return hash(self._branches[branch])

    def has(self, name: str, branch: Branch | str | None = None) -> bool:
        try:
            self.get(name=name, branch=branch, duplicate=False)
            return True
        except SchemaNotFoundError:
            return False

    def get(
        self,
        name: str,
        branch: Branch | str | None = None,
        duplicate: bool = True,
        check_branch_only: bool = False,
    ) -> MainSchemaTypes:
        # For now we assume that all branches are present, will see how we need to pull new branches later.
        check_branch_only = check_branch_only and bool(branch)
        branch = registry.get_branch_from_registry(branch=branch)

        if branch.name in self._branches:
            try:
                return self._branches[branch.name].get(name=name, duplicate=duplicate)
            except SchemaNotFoundError:
                pass

        if check_branch_only:
            raise SchemaNotFoundError(
                branch_name=branch.name, identifier=name, message=f"Unable to find the schema {name!r} in the registry"
            )

        default_branch = registry.default_branch
        return self._branches[default_branch].get(name=name, duplicate=duplicate)

    def get_node_schema(self, name: str, branch: Branch | str | None = None, duplicate: bool = True) -> NodeSchema:
        schema = self.get(name=name, branch=branch, duplicate=duplicate)
        if isinstance(schema, NodeSchema):
            return schema

        raise ValueError("The selected node is not of type NodeSchema")

    def get_profile_schema(
        self, name: str, branch: Branch | str | None = None, duplicate: bool = True
    ) -> ProfileSchema:
        schema = self.get(name=name, branch=branch, duplicate=duplicate)
        if isinstance(schema, ProfileSchema):
            return schema

        raise ValueError("The selected node is not of type ProfileSchema")

    def get_template_schema(
        self, name: str, branch: Branch | str | None = None, duplicate: bool = True
    ) -> TemplateSchema:
        schema = self.get(name=name, branch=branch, duplicate=duplicate)
        if isinstance(schema, TemplateSchema):
            return schema

        raise ValueError("The selected node is not of type TemplateSchema")

    def get_full(self, branch: Branch | str | None = None, duplicate: bool = True) -> dict[str, MainSchemaTypes]:
        branch = registry.get_branch_from_registry(branch=branch)

        branch_name = None
        if branch.name in self._branches:
            branch_name = branch.name
        else:
            branch_name = registry.default_branch

        return self._branches[branch_name].get_all(duplicate=duplicate)

    async def get_full_safe(self, branch: Branch | str | None = None) -> dict[str, NodeSchema | GenericSchema]:
        await lock.registry.local_schema_wait()

        return self.get_full(branch=branch)

    def get_schema_branch(self, name: str) -> SchemaBranch:
        if name in self._branches:
            return self._branches[name]

        self._branches[name] = SchemaBranch(cache=self._cache, name=name)
        return self._branches[name]

    def set_schema_branch(self, name: str, schema: SchemaBranch) -> None:
        schema.name = name
        self._branches[name] = schema

    def process_schema_branch(self, name: str) -> None:
        schema_branch = self.get_schema_branch(name=name)
        schema_branch.process()

    async def update_schema_branch(
        self,
        schema: SchemaBranch,
        db: InfrahubDatabase,
        branch: Branch | str | None = None,
        diff: SchemaDiff | None = None,
        limit: list[str] | None = None,
        update_db: bool = True,
    ) -> None:
        branch = await registry.get_branch(branch=branch, db=db)

        updated_schema = None
        if update_db:
            if diff:
                schema_diff = await self.update_schema_to_db(schema=schema, db=db, branch=branch, diff=diff)
            else:
                await self.load_schema_to_db(schema=schema, db=db, branch=branch, limit=limit)
                # After updating the schema into the db
                # we need to pull a fresh version because some default value are managed/generated within the node object
                schema_diff = None
                if limit:
                    schema_diff = SchemaBranchDiff(
                        added_nodes=[name for name in list(schema.nodes.keys()) if name in limit],
                        added_generics=[name for name in list(schema.generics.keys()) if name in limit],
                    )

            updated_schema = await self.load_schema_from_db(
                db=db, branch=branch, schema=schema, schema_diff=schema_diff
            )

        self.set_schema_branch(name=branch.name, schema=updated_schema or schema)

    def register_schema(self, schema: SchemaRoot, branch: str | None = None) -> SchemaBranch:
        """Register all nodes, generics & groups from a SchemaRoot object into the registry."""

        branch = branch or registry.default_branch
        schema_branch = self.get_schema_branch(name=branch)
        schema_branch.load_schema(schema=schema)
        schema_branch.process()
        return schema_branch

    async def update_schema_to_db(
        self,
        schema: SchemaBranch,
        db: InfrahubDatabase,
        diff: SchemaDiff,
        branch: Branch | str | None = None,
    ) -> SchemaBranchDiff:
        """Load all nodes, generics and groups from a SchemaRoot object into the database."""

        branch = await registry.get_branch(branch=branch, db=db)

        added_nodes = []
        added_generics = []
        for item_kind in diff.added.keys():
            item = schema.get(name=item_kind, duplicate=False)
            node = await self.load_node_to_db(node=item, branch=branch, db=db)
            schema.set(name=item_kind, schema=node)
            if item.is_node_schema:
                added_nodes.append(item_kind)
            else:
                added_generics.append(item_kind)

        changed_nodes = []
        changed_generics = []
        for item_kind, item_diff in diff.changed.items():
            item = schema.get(name=item_kind, duplicate=False)
            if item_diff:
                node = await self.update_node_in_db_based_on_diff(node=item, branch=branch, db=db, diff=item_diff)
            else:
                node = await self.update_node_in_db(node=item, branch=branch, db=db)
            schema.set(name=item_kind, schema=node)
            if item.is_node_schema:
                changed_nodes.append(item_kind)
            else:
                changed_generics.append(item_kind)

        removed_nodes = []
        removed_generics = []
        for item_kind in diff.removed.keys():
            item = schema.get(name=item_kind, duplicate=False)
            node = await self.delete_node_in_db(node=item, branch=branch, db=db)
            schema.delete(name=item_kind)
            if item.is_node_schema:
                removed_nodes.append(item_kind)
            else:
                removed_generics.append(item_kind)

        return SchemaBranchDiff(
            added_nodes=added_nodes,
            added_generics=added_generics,
            changed_nodes=changed_nodes,
            changed_generics=changed_generics,
            removed_nodes=removed_nodes,
            removed_generics=removed_generics,
        )

    async def load_schema_to_db(
        self,
        schema: SchemaBranch,
        db: InfrahubDatabase,
        branch: Branch | str | None = None,
        limit: list[str] | None = None,
    ) -> None:
        """Load all nodes, generics and groups from a SchemaRoot object into the database."""

        branch = await registry.get_branch(branch=branch, db=db)

        for item_kind in schema.node_names + schema.generic_names_without_templates:
            if limit and item_kind not in limit:
                continue
            item = schema.get(name=item_kind, duplicate=False)
            if not item.id:
                node = await self.load_node_to_db(node=item, branch=branch, db=db)
                schema.set(name=item_kind, schema=node)
            else:
                node = await self.update_node_in_db(node=item, branch=branch, db=db)
                schema.set(name=item_kind, schema=node)

    async def load_node_to_db(
        self,
        node: NodeSchema | GenericSchema,
        db: InfrahubDatabase,
        branch: Branch | str | None = None,
    ) -> NodeSchema | GenericSchema:
        """Load a Node with its attributes and its relationships to the database."""
        branch = await registry.get_branch(branch=branch, db=db)

        node_type = "SchemaNode"
        if isinstance(node, GenericSchema):
            node_type = "SchemaGeneric"

        node_schema = self.get_node_schema(name=node_type, branch=branch, duplicate=False)
        attribute_schema = self.get_node_schema(name="SchemaAttribute", branch=branch, duplicate=False)
        relationship_schema = self.get_node_schema(name="SchemaRelationship", branch=branch, duplicate=False)

        # Duplicate the node in order to store the IDs after inserting them in the database
        new_node = node.duplicate()

        # Create the node first
        schema_dict = node.model_dump(exclude={"id", "state", "filters", "relationships", "attributes"})
        obj = await Node.init(schema=node_schema, branch=branch, db=db)
        await obj.new(**schema_dict, db=db)
        await obj.save(db=db)
        new_node.id = obj.id

        # Then create the Attributes and the relationships
        if isinstance(node, NodeSchema | GenericSchema):
            new_node.relationships = []
            new_node.attributes = []

            for item in node.attributes:
                if item.inherited is False:
                    new_attr = await self.create_attribute_in_db(
                        schema=attribute_schema, item=item, parent=obj, branch=branch, db=db
                    )
                else:
                    new_attr = item.duplicate()
                new_node.attributes.append(new_attr)

            for item in node.relationships:
                if item.inherited is False:
                    new_rel = await self.create_relationship_in_db(
                        schema=relationship_schema, item=item, parent=obj, branch=branch, db=db
                    )
                else:
                    new_rel = item.duplicate()
                new_node.relationships.append(new_rel)

        # Save back the node with the newly created IDs in the SchemaManager
        self.set(name=new_node.kind, schema=new_node, branch=branch.name)
        return new_node

    async def update_node_in_db(
        self,
        db: InfrahubDatabase,
        node: NodeSchema | GenericSchema,
        branch: Branch | str | None = None,
    ) -> NodeSchema | GenericSchema:
        """Update a Node with its attributes and its relationships in the database."""
        branch = await registry.get_branch(branch=branch, db=db)

        obj = await self.get_one(id=node.get_id(), branch=branch, db=db)
        if not obj:
            raise SchemaNotFoundError(
                branch_name=branch.name,
                identifier=node.id,
                message=f"Unable to find the Schema associated with {node.id}, {node.kind}",
            )

        schema_dict = node.model_dump(exclude=IGNORE_FOR_NODE)
        for key, value in schema_dict.items():
            getattr(obj, key).value = value

        attribute_schema = self.get_node_schema(name="SchemaAttribute", branch=branch)
        relationship_schema = self.get_node_schema(name="SchemaRelationship", branch=branch)

        new_node = node.duplicate()

        # Update the attributes and the relationships nodes as well
        await obj.attributes.update(db=db, data=[item.id for item in node.local_attributes if item.id])
        await obj.relationships.update(
            db=db, data=[item.id for item in node.local_relationships if item.id and item.name != "profiles"]
        )
        await obj.save(db=db)

        # Then Update the Attributes and the relationships

        items = await self.get_many(
            ids=[item.id for item in node.local_attributes + node.local_relationships if item.id],
            db=db,
            branch=branch,
            include_owner=True,
            include_source=True,
        )

        for item in node.local_attributes:
            if item.id and item.id in items:
                await self.update_attribute_in_db(item=item, attr=items[item.id], db=db)
            elif not item.id:
                new_attr = await self.create_attribute_in_db(
                    schema=attribute_schema, item=item, branch=branch, db=db, parent=obj
                )
                new_node.attributes.append(new_attr)

        for item in node.local_relationships:
            if item.id and item.id in items:
                await self.update_relationship_in_db(item=item, rel=items[item.id], db=db)
            elif not item.id:
                new_rel = await self.create_relationship_in_db(
                    schema=relationship_schema, item=item, branch=branch, db=db, parent=obj
                )
                new_node.relationships.append(new_rel)

        # Save back the node with the (potentially) newly created IDs in the SchemaManager
        self.set(name=new_node.kind, schema=new_node, branch=branch.name)
        return new_node

    async def update_node_in_db_based_on_diff(
        self,
        db: InfrahubDatabase,
        diff: HashableModelDiff,
        node: NodeSchema | GenericSchema,
        branch: Branch | str | None = None,
    ) -> NodeSchema | GenericSchema:
        """Update a Node with its attributes and its relationships in the database based on a HashableModelDiff."""
        branch = await registry.get_branch(branch=branch, db=db)

        obj = await self.get_one(id=node.get_id(), branch=branch, db=db)
        if not obj:
            raise SchemaNotFoundError(
                branch_name=branch.name,
                identifier=node.id,
                message=f"Unable to find the Schema associated with {node.id}, {node.kind}",
            )

        properties_to_update = set(list(diff.added.keys()) + list(diff.changed.keys())) - IGNORE_FOR_NODE

        if properties_to_update:
            schema_dict = node.model_dump(exclude=IGNORE_FOR_NODE)
            for key, value in schema_dict.items():
                getattr(obj, key).value = value

        attribute_schema = self.get_node_schema(name="SchemaAttribute", branch=branch)
        relationship_schema = self.get_node_schema(name="SchemaRelationship", branch=branch)

        new_node = node.duplicate()

        diff_attributes = diff.changed.get("attributes")
        diff_relationships = diff.changed.get("relationships")
        attrs_rels_to_update: set[str] = set()
        if diff_attributes:
            attrs_rels_to_update.update(set(diff_attributes.added.keys()))
            attrs_rels_to_update.update(set(diff_attributes.changed.keys()))
            attrs_rels_to_update.update(set(diff_attributes.removed.keys()))
        if diff_relationships:
            attrs_rels_to_update.update(set(diff_relationships.added.keys()))
            attrs_rels_to_update.update(set(diff_relationships.changed.keys()))
            attrs_rels_to_update.update(set(diff_relationships.removed.keys()))

        item_ids = set()
        item_names = set()
        for field in node.local_attributes + node.local_relationships:
            if field.name not in attrs_rels_to_update:
                continue
            if field.id:
                item_ids.add(field.id)
                item_names.add(field.name)
        missing_field_names = list(attrs_rels_to_update - item_names)

        items: dict[str, Node] = {}
        if item_ids:
            items = await self.get_many(
                ids=list(item_ids),
                db=db,
                branch=branch,
                include_owner=True,
                include_source=True,
            )
        if missing_field_names:
            missing_attrs = await self.query(
                db=db,
                branch=branch,
                schema=attribute_schema,
                filters={"name__values": missing_field_names, "node__id": node.id},
                include_owner=True,
                include_source=True,
            )
            missing_rels = await self.query(
                db=db,
                branch=branch,
                schema=relationship_schema,
                filters={"name__values": missing_field_names, "node__id": node.id},
                include_owner=True,
                include_source=True,
            )
            items.update({field.id: field for field in missing_attrs + missing_rels})

        if diff_attributes:
            await obj.attributes.update(db=db, data=[item.id for item in node.local_attributes if item.id])

        if diff_relationships:
            await obj.relationships.update(db=db, data=[item.id for item in node.local_relationships if item.id])

        await obj.save(db=db)

        if diff_attributes:
            for item in node.local_attributes:
                # if item is in changed and has no ID, then it is being overridden from a generic and must be added
                if item.name in diff_attributes.added or (item.name in diff_attributes.changed and item.id is None):
                    created_item = await self.create_attribute_in_db(
                        schema=attribute_schema, item=item, branch=branch, db=db, parent=obj
                    )
                    new_attr = new_node.get_attribute(name=item.name)
                    new_attr.id = created_item.id
                elif item.name in diff_attributes.changed and item.id and item.id in items:
                    await self.update_attribute_in_db(item=item, attr=items[item.id], db=db)
                elif item.name in diff_attributes.removed and item.id and item.id in items:
                    await items[item.id].delete(db=db)
                elif (
                    (item.name in diff_attributes.removed or item.name in diff_attributes.changed)
                    and item.id
                    and item.id not in items
                ):
                    raise ValueError(f"Unable to find an attribute {item.name!r} to update or delete")

        if diff_relationships:
            for item in node.local_relationships:
                # if item is in changed and has no ID, then it is being overridden from a generic and must be added
                if item.name in diff_relationships.added or (
                    item.name in diff_relationships.changed and item.id is None
                ):
                    created_rel = await self.create_relationship_in_db(
                        schema=relationship_schema, item=item, branch=branch, db=db, parent=obj
                    )
                    new_rel = new_node.get_relationship(name=item.name)
                    new_rel.id = created_rel.id
                elif item.name in diff_relationships.changed and item.id and item.id in items:
                    await self.update_relationship_in_db(item=item, rel=items[item.id], db=db)
                elif item.name in diff_relationships.removed and item.id and item.id in items:
                    await items[item.id].delete(db=db)
                elif (
                    (item.name in diff_relationships.removed or item.name in diff_relationships.changed)
                    and item.id
                    and item.id not in items
                ):
                    raise ValueError(f"Unable to find a relationship {item.name!r} to update or delete")

        field_names_to_remove = []
        if diff_attributes and diff_attributes.removed:
            attr_names_to_remove = set(diff_attributes.removed.keys()) - set(node.local_attribute_names)
            field_names_to_remove.extend(list(attr_names_to_remove))
        if diff_relationships and diff_relationships.removed:
            rel_names_to_remove = set(diff_relationships.removed.keys()) - set(node.local_relationship_names)
            field_names_to_remove.extend(list(rel_names_to_remove))
        if field_names_to_remove:
            for field_schema in items.values():
                if field_schema.name.value in field_names_to_remove:
                    await field_schema.delete(db=db)

        # Save back the node with the (potentially) newly created IDs in the SchemaManager
        self.set(name=new_node.kind, schema=new_node, branch=branch.name)
        return new_node

    async def delete_node_in_db(
        self,
        db: InfrahubDatabase,
        node: NodeSchema | GenericSchema,
        branch: Branch | str | None = None,
    ) -> None:
        """Delete the node with its attributes and relationships."""
        branch = await registry.get_branch(branch=branch, db=db)

        obj = await self.get_one(id=node.get_id(), branch=branch, db=db, prefetch_relationships=True)
        if not obj:
            raise SchemaNotFoundError(
                branch_name=branch.name,
                identifier=node.id,
                message=f"Unable to find the Schema associated with {node.id}, {node.kind}",
            )

        # First delete the attributes and the relationships
        for attr_schema_node in (await obj.attributes.get_peers(db=db)).values():
            await attr_schema_node.delete(db=db)
        for rel_schema_node in (await obj.relationships.get_peers(db=db)).values():
            await rel_schema_node.delete(db=db)

        await obj.delete(db=db)

    @staticmethod
    async def create_attribute_in_db(
        schema: NodeSchema, item: AttributeSchema, branch: Branch, parent: Node, db: InfrahubDatabase
    ) -> AttributeSchema:
        obj = await Node.init(schema=schema, branch=branch, db=db)
        await obj.new(**item.to_node(), node=parent, db=db)
        await obj.save(db=db)
        new_item = item.duplicate()
        new_item.id = obj.id
        return new_item

    @staticmethod
    async def update_attribute_in_db(item: AttributeSchema, attr: Node, db: InfrahubDatabase) -> None:
        item_dict = item.model_dump(exclude={"id", "state", "filters"})
        for key, value in item_dict.items():
            getattr(attr, key).value = value
        await attr.save(db=db)

    @staticmethod
    async def create_relationship_in_db(
        schema: NodeSchema, item: RelationshipSchema, branch: Branch, parent: Node, db: InfrahubDatabase
    ) -> RelationshipSchema:
        obj = await Node.init(schema=schema, branch=branch, db=db)
        await obj.new(**item.model_dump(exclude={"id", "state", "filters"}), node=parent, db=db)
        await obj.save(db=db)
        new_item = item.duplicate()
        new_item.id = obj.id
        return new_item

    @staticmethod
    async def update_relationship_in_db(item: RelationshipSchema, rel: Node, db: InfrahubDatabase) -> None:
        item_dict = item.model_dump(exclude={"id", "state", "filters"})
        for key, value in item_dict.items():
            getattr(rel, key).value = value
        await rel.save(db=db)

    async def load_schema(
        self,
        db: InfrahubDatabase,
        branch: Branch | str | None = None,
    ) -> SchemaBranch:
        """Load the schema either from the cache or from the database"""
        branch = await registry.get_branch(branch=branch, db=db)

        if not branch.is_default and branch.origin_branch:
            origin_branch: Branch = await registry.get_branch(branch=branch.origin_branch, db=db)

            if origin_branch.active_schema_hash.main == branch.active_schema_hash.main:
                origin_schema = self.get_schema_branch(name=origin_branch.name)
                new_branch_schema = origin_schema.duplicate()
                self.set_schema_branch(name=branch.name, schema=new_branch_schema)
                log.info("Loading schema from cache")
                return new_branch_schema

        current_schema = self.get_schema_branch(name=branch.name)
        schema_diff = current_schema.get_hash_full().compare(branch.active_schema_hash)
        branch_schema = await self.load_schema_from_db(
            db=db, branch=branch, schema=current_schema, schema_diff=schema_diff
        )
        self.set_schema_branch(name=branch.name, schema=branch_schema)
        return branch_schema

    async def load_schema_from_db(
        self,
        db: InfrahubDatabase,
        branch: Branch | str | None = None,
        schema: SchemaBranch | None = None,
        schema_diff: SchemaBranchDiff | None = None,
        at: Timestamp | None = None,
        validate_schema: bool = True,
    ) -> SchemaBranch:
        """Query all the node of type NodeSchema and GenericSchema from the database and convert them to their respective type.

        Args:
            db: Database Driver
            branch: Name of the branch to load the schema from. Defaults to None.
            schema: (Optional) If a schema is provided, it will be updated with the latest value, if not a new one will be created.
            schema_diff: (Optional). list of nodes, generics & groups to query

        Returns:
            SchemaBranch
        """

        branch = await registry.get_branch(branch=branch, db=db)
        schema = schema or SchemaBranch(cache=self._cache, name=branch.name)

        # If schema_diff has been provided, we need to build the proper filters for the queries based on the namespace and the name of the object.
        # the namespace and the name will be extracted from the kind with the function `parse_node_kind`
        filters = {"generics": {}, "nodes": {}}
        has_filters = False

        # If a diff is provided but is empty there is nothing to query
        if schema_diff is not None and not schema_diff.has_diff:
            return schema

        if schema_diff:
            log.info("Loading schema from DB", schema_to_update=schema_diff.to_list())

            for node_type in list(filters.keys()):
                filter_value = {
                    "namespace__values": list(
                        {parse_node_kind(item).namespace for item in getattr(schema_diff, node_type)}
                    ),
                    "name__values": list({parse_node_kind(item).name for item in getattr(schema_diff, node_type)}),
                }

                if filter_value["namespace__values"]:
                    filters[node_type] = filter_value
                    has_filters = True
            for removed_generic in schema_diff.removed_generics:
                if removed_generic in schema.generic_names:
                    schema.delete(name=removed_generic)
            for removed_node in schema_diff.removed_nodes:
                if removed_node in schema.node_names:
                    schema.delete(name=removed_node)

        if not has_filters or filters["generics"]:
            generic_schema = self.get(name="SchemaGeneric", branch=branch)
            for schema_node in await self.query(
                schema=generic_schema,
                branch=branch,
                at=at,
                filters=filters["generics"],
                prefetch_relationships=True,
                db=db,
            ):
                kind = f"{schema_node.namespace.value}{schema_node.name.value}"
                schema.set(
                    name=kind,
                    schema=await self.convert_generic_schema_to_schema(schema_node=schema_node, db=db),
                )

        if not has_filters or filters["nodes"]:
            node_schema = self.get(name="SchemaNode", branch=branch)
            for schema_node in await self.query(
                schema=node_schema, branch=branch, at=at, filters=filters["nodes"], prefetch_relationships=True, db=db
            ):
                kind = f"{schema_node.namespace.value}{schema_node.name.value}"
                schema.set(
                    name=kind,
                    schema=await self.convert_node_schema_to_schema(schema_node=schema_node, db=db),
                )

        schema.process(validate_schema=validate_schema)

        return schema

    @classmethod
    async def _prepare_node_data(cls, schema_node: Node, db: InfrahubDatabase) -> dict[str, Any]:
        node_data = {"id": schema_node.id}

        # First pull all the local attributes at the top level, then convert all the local relationships
        #  for a standard node_schema, the relationships will be attributes and relationships
        for attr_name in schema_node._attributes:
            attr = getattr(schema_node, attr_name)
            node_data[attr_name] = attr.get_value()

        for rel_name in schema_node._relationships:
            if rel_name not in node_data:
                if rel_name == "profiles":
                    continue
                node_data[rel_name] = []

            rm = getattr(schema_node, rel_name)
            for rel in await rm.get(db=db):
                item = await rel.get_peer(db=db)
                item_data = {"id": item.id}
                inherited_attr = getattr(item, "inherited", None)
                if inherited_attr and getattr(inherited_attr, "value", False) is True:
                    continue
                for item_name in item._attributes:
                    item_attr = getattr(item, item_name)
                    item_data[item_name] = item_attr.get_value()

                node_data[rel_name].append(item_data)
        return node_data

    @classmethod
    async def convert_node_schema_to_schema(cls, schema_node: Node, db: InfrahubDatabase) -> NodeSchema:
        """Convert a schema_node object loaded from the database into NodeSchema object."""
        node_data = await cls._prepare_node_data(schema_node=schema_node, db=db)
        return NodeSchema(**node_data)

    @classmethod
    async def convert_generic_schema_to_schema(cls, schema_node: Node, db: InfrahubDatabase) -> GenericSchema:
        """Convert a schema_node object loaded from the database into GenericSchema object."""
        node_data = await cls._prepare_node_data(schema_node=schema_node, db=db)
        return GenericSchema(**node_data)

    def purge_inactive_branches(self, active_branches: list[str]) -> list[str]:
        """Return non active branches that were purged."""

        hashes_to_keep: set[str] = set()
        for active_branch in active_branches:
            if branch := self._branches.get(active_branch):
                nodes = branch.get_all(include_internal=True, duplicate=False)
                hashes_to_keep.update([node.get_hash() for node in nodes.values()])

        removed_branches: list[str] = []
        for branch_name in list(self._branches.keys()):
            if branch_name not in active_branches:
                del self._branches[branch_name]
                removed_branches.append(branch_name)

        for hash_key in list(self._cache.keys()):
            if hash_key not in hashes_to_keep:
                del self._cache[hash_key]

        return removed_branches

    def get_branches(self) -> list[str]:
        return list(self._branches.keys())



from infrahub.core.diff.enricher.aggregated import AggregatedDiffEnricher
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from .cardinality_one import DiffCardinalityOneEnricherDependency
from .hierarchy import DiffHierarchyEnricherDependency
from .labels import DiffLabelsEnricherDependency
from .path_identifier import DiffPathIdentifierEnricherDependency


class DiffAggregatedEnricherDependency(DependencyBuilder[AggregatedDiffEnricher]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AggregatedDiffEnricher:
        return AggregatedDiffEnricher(
            enrichers=[
                DiffCardinalityOneEnricherDependency.build(context=context),
                DiffHierarchyEnricherDependency.build(context=context),
                DiffLabelsEnricherDependency.build(context=context),
                DiffPathIdentifierEnricherDependency.build(context=context),
            ]
        )
